'''
Created on Aug 1, 2013

@author: sean
'''

from __future__ import (print_function, unicode_literals, division,
    absolute_import)

from binstar_client.utils import jencode, compute_hash
from binstar_client.requests_ext import stream_multipart
import requests
from binstar_client.errors import BinstarError

class BuildMixin(object):
    '''
    Add build functionality to binstar client
    '''

    def set_keyfile(self, username, package, filename, content):
        url = '%s/build/%s/%s/keyfile' % (self.domain, username, package)
        data, headers = jencode(filename=filename, content=content)
        res = self.session.post(url, data=data, headers=headers)
        self._check_response(res, [201])
    def remove_keyfile(self, username, package, filename):
        url = '%s/build/%s/%s/keyfile' % (self.domain, username, package)
        params = dict(filename=filename)
        res = self.session.delete(url, params=params)
        self._check_response(res, [201])

    def keyfiles(self, username, package):
        url = '%s/build/%s/%s/keyfiles' % (self.domain, username, package)
        res = self.session.get(url)
        self._check_response(res)
        return res.json()

    def submit_for_build(self, username, package, fd, instructions,
                         test_only=False, channels=None, queue=None, queue_tags=None, callback=None):

        url = '%s/build/%s/%s/stage' % (self.domain, username, package)
        data, headers = jencode(instructions=instructions, test_only=test_only, channels=channels, queue_name=queue, queue_tags=queue_tags)

        res = self.session.post(url, data=data, headers=headers)
        self._check_response(res)
        obj = res.json()

        s3url = obj['post_url']
        s3data = obj['form_data']

        _hexmd5, b64md5, size = compute_hash(fd)
        s3data['Content-Length'] = size
        s3data['Content-MD5'] = b64md5

        data_stream, headers = stream_multipart(s3data, files={'file':(obj['basename'], fd)},
                                                callback=callback)

        s3res = requests.post(s3url, data=data_stream, verify=True, timeout=10 * 60 * 60, headers=headers)

        if s3res.status_code != 201:
            raise BinstarError('Error uploading build', s3res.status_code)

        url = '%s/build/%s/%s/commit/%s' % (self.domain, username, package, obj['build_id'])
        res = self.session.post(url, verify=True)
        self._check_response(res, [201])
        return obj

    def submit_for_url_build(self, username, package, instructions,
                             test_only=False, callback=None,
                             channels=None, queue=None, queue_tags=None, sub_dir='',
                             filter_platform=None):

        # /build/<owner_login>/<package_name>/submit-git-url
        url = '%s/build/%s/%s/submit-git-url' % (self.domain, username, package)

        data, headers = jencode(instructions=instructions, test_only=test_only,
                                channels=channels, sub_dir=sub_dir,
                                queue_name=queue, queue_tags=queue_tags,
                                filter_platform=filter_platform)

        res = self.session.post(url, data=data, headers=headers)

        self._check_response(res, [201])
        obj = res.json()
        return obj

    def builds(self, username, package, build_no=None):
        if build_no:
            url = '%s/build/%s/%s/%s' % (self.domain, username, package, build_no)
        else:
            url = '%s/build/%s/%s' % (self.domain, username, package)
        res = self.session.get(url)
        self._check_response(res)
        return res.json()

    def stop_build(self, username, package, build_id):
        url = '%s/build/%s/%s/stop/%s' % (self.domain, username, package, build_id)
        res = self.session.post(url)
        self._check_response(res, [201])
        return

    def tail_build(self, username, package, build_id, limit='', after=''):
        url = '%s/build/%s/%s/tail/%s' % (self.domain, username, package, build_id)
        res = self.session.get(url, params={'limit':limit, 'after': after})
        self._check_response(res, [200])
        return res.json()

    def resubmit_build(self, username, package, build_id):
        url = '%s/build/%s/%s/resubmit/%s' % (self.domain, username, package, build_id)
        res = self.session.post(url)
        self._check_response(res, [201])
        return

    def add_ci(self, username, package,
                   ghowner, ghrepo,
                   channels, queue=None, sub_dir=None,
                   branch='master', email=None):

        url = '%s/build/%s/%s/ci' % (self.domain, username, package)
        data, headers = jencode(ghowner=ghowner, ghrepo=ghrepo,
                                channels=channels, queue_name=queue, sub_dir=sub_dir,
                                branch=branch,
                                email=email,
                                )
        res = self.session.post(url, data=data, headers=headers)
        self._check_response(res, [201])

    def remove_ci(self, username, package):
        url = '%s/build/%s/%s/ci' % (self.domain, username, package)
        res = self.session.delete(url)
        self._check_response(res, [201])

    def trigger_build(self, username, package, channels=None,
                      queue_name=None, branch=None, test_only=False,
                      filter_platform=None):
        url = '%s/build/%s/%s/trigger' % (self.domain, username, package)

        data, headers = jencode(channels=channels, queue_name=queue_name,
                                branch=branch, test_only=test_only,
                                filter_platform=filter_platform)

        res = self.session.post(url, data=data, headers=headers)
        self._check_response(res, [201])

        obj = res.json()
        return obj['build_no']


    def upload_test_results(self, action, username, package, major, minor, fd):
        url = '%s/build/%s/%s/results/%i/%i/%s' % (self.domain, username, package, major, minor, action)
        headers = {'Content-Type': 'application/xml'}
        res = self.session.post(url, data=fd.read(), headers=headers)
        self._check_response(res)

        return




'''
Created on Sep 23, 2013

@author: sean
'''


from __future__ import (print_function, unicode_literals, division,
    absolute_import)

from coverage.summary import SummaryReporter
from coverage.results import Numbers
from coverage.misc import NotPython
import sys

WARNING = '\033[33m'
OKBLUE = '\033[34m'
OKGREEN = '\033[32m'
FAIL = '\033[31m'
ENDC = '\033[0m'
BOLD = "\033[1m"

def green(text):
    return BOLD + OKGREEN + text + ENDC

def red(text):
    return FAIL + text + ENDC

def orange(text):
    return WARNING + text + ENDC

def blue(text):
    return OKBLUE + text + ENDC

class ColorSummaryReporter(SummaryReporter):

    @property
    def fmt_name(self):
        max_name = max([len(cu.name) for cu in self.code_units] + [5])
        fmt_name = "%%- %ds  " % max_name
        return fmt_name

    def fmt_coverage(self, perc):

        fmt_coverage = self.fmt_name + "%6d %6d"
        if self.branches:
            fmt_coverage += " %6d %6d"
        width100 = Numbers.pc_str_width()
        color = green
        if perc < 80:
            color = blue
        if perc < 50:
            color = red

        fmt_coverage += color("%%%ds%%%%" % (width100+3,))
        if self.config.show_missing:
            fmt_coverage += "   %s"
        fmt_coverage += "\n"
        return fmt_coverage

    def header(self):
        header = (self.fmt_name % "Name") + " Stmts   Miss"
        if self.branches:
            header += " Branch BrMiss"
        width100 = Numbers.pc_str_width()
        header += "%*s" % (width100+4, "Cover")
        header += "\n"
        return header

    def report(self, morfs, outfile=None):
        """Writes a report summarizing coverage statistics per module.

        `outfile` is a file object to write the summary to.

        """
        self.find_code_units(morfs)

        # Prepare the formatting strings

        fmt_err = "%s   %s: %s\n"

        header = self.header()
#         fmt_coverage = self.fmt_coverage()
        rule = "-" * len(header) + "\n"

        if not outfile:
            outfile = sys.stdout

        # Write the header
        outfile.write(header)
        outfile.write(rule)

        total = Numbers()

        for cu in self.code_units:
            try:
                analysis = self.coverage._analyze(cu)
                nums = analysis.numbers
                args = (cu.name, nums.n_statements, nums.n_missing)
                if self.branches:
                    args += (nums.n_branches, nums.n_missing_branches)
                args += (nums.pc_covered_str,)
                if self.config.show_missing:
                    args += (analysis.missing_formatted(),)
                outfile.write(self.fmt_coverage(nums.pc_covered) % args)
                total += nums
            except KeyboardInterrupt:                   # pragma: not covered
                raise
            except:
                report_it = not self.config.ignore_errors
                if report_it:
                    typ, msg = sys.exc_info()[:2]
                    if typ is NotPython and not cu.should_be_python():
                        report_it = False
                if report_it:
                    outfile.write(fmt_err % (cu.name, typ.__name__, msg))

        if total.n_files > 1:
            outfile.write(rule)
            args = ("TOTAL", total.n_statements, total.n_missing)
            if self.branches:
                args += (total.n_branches, total.n_missing_branches)
            args += (total.pc_covered_str,)
            if self.config.show_missing:
                args += ("",)
#             outfile.write(self.fmt_coverage(total.pc_covered) % args)
        return total


def report(cov, morfs=None, show_missing=False, ignore_errors=None,
                file=None,                          # pylint: disable=W0622
                omit=None, include=None
                ):
        cov._harvest_data()
        cov.config.from_args(
            ignore_errors=ignore_errors, omit=omit, include=include,
            show_missing=show_missing,
            )
        reporter = ColorSummaryReporter(cov, cov.config)
        return reporter.report(morfs, outfile=file)



'''

'''

from argparse import RawDescriptionHelpFormatter, FileType
from binstar_client.utils import package_specs, get_binstar
from binstar_build_client import BinstarBuildAPI

def main(args):
    bs = get_binstar(args, BinstarBuildAPI)

    if  '.' in args.build:
        major, minor = args.build.split('.', 1)
    else:
        major, minor = args.build, 0

    major, minor = int(major), int(minor)

    bs.upload_test_results(args.action, args.package.user, args.package.name,
                           major, minor, args.filename,
                           )

def add_parser(subparsers):

    description = '[Advanced] Attach results to build'
    parser = subparsers.add_parser('results',
                                      help=description,
                                      description=description,
                                      epilog=__doc__,
                                      formatter_class=RawDescriptionHelpFormatter
                                      )
    # ArgumentParser.add_subparsers(self)
    parser.add_argument('action',
                   choices=['testsuite', 'summary'])

    parser.add_argument('package',
                   metavar='USER/PACKAGE',
                   type=package_specs)
    parser.add_argument('build', metavar='MAJOR.MINOR')
    parser.add_argument('filename', type=FileType('r'))

    parser.set_defaults(main=main)



'''
Created on Aug 5, 2013

@author: sean
'''

from __future__ import (print_function, unicode_literals, division,
    absolute_import)

import unittest
from os.path import dirname,join
from argparse import ArgumentParser
from binstar_client.tests.runner import ColorTextTestRunner
from binstar_client.tests.coverage_report import report

def main():

    parser = ArgumentParser()
    parser.add_argument('--html', action='store_true')
    parser.add_argument('source_dir', nargs='?', default='')
    args = parser.parse_args()

    import coverage
    cov = coverage.coverage(include='**%s**' % args.source_dir if args.source_dir else '**/binstar_client/**',
                            omit=['**/lib/python2.7/**', '**/site-packages/**', '**/tests/**',
                                  ])

    cov.start()
    import binstar_client
    print(binstar_client)
    loader = unittest.loader.TestLoader()
    discover_dir = join(dirname(binstar_client.__path__[0]), args.source_dir)
    print('Discover %s' % discover_dir)
    tests = loader.discover(discover_dir)
    runner = ColorTextTestRunner(verbosity=2)
    result = runner.run(tests)
    cov.stop()
    cov.save()
    total = report(cov)
    if args.html:
        cov.html_report(directory='htmlcov')

    runner.write_end(result, total)

    exit(0 if result.wasSuccessful() else -1)


if __name__ == '__main__':
    main()



import unittest
from binstar_build_client.worker.utils.buffered_io import BufferedPopen
import io
import os
skipWindows = unittest.skipIf(os.name == 'nt', 'This test does not run on windows')
skipPosix = unittest.skipIf(os.name != 'nt', 'This test does not run on Posix')

class Test(unittest.TestCase):


    def test_stdout(self):
        stdout = io.BytesIO()
        p0 = BufferedPopen(['echo', 'hello'], stdout=stdout)
        self.assertEqual(p0.wait(), 0)
        self.assertEqual(stdout.getvalue().strip(), 'hello')

    @skipWindows
    def test_stdout_error(self):
        stdout = io.BytesIO()
        p0 = BufferedPopen(['bash', '-c', 'echo hello; bad-command'], stdout=stdout)
        return_code = p0.wait()
        self.assertNotEqual(return_code, 0)
        self.assertTrue(stdout.getvalue().startswith('hello\n'))

    @skipPosix
    def test_stdout_error_win(self):
        stdout = io.BytesIO()
        p0 = BufferedPopen(['cmd', '/c', 'echo hello & bad-command'], stdout=stdout)
        return_code = p0.wait()
        self.assertNotEqual(return_code, 0)

        self.assertTrue(stdout.getvalue().startswith('hello'))

    @skipWindows
    def test_stderr(self):
        stdout = io.BytesIO()
        p0 = BufferedPopen(['bash', '-c', 'echo stdout 2>&1; echo stderr 1>&2;'], stdout=stdout)
        self.assertEqual(p0.wait(), 0)
        self.assertEqual(stdout.getvalue(), 'stdout\nstderr\n')

    @skipWindows
    def test_iotimeout(self):
        stdout = io.BytesIO()
        p0 = BufferedPopen(['bash', '-c', 'echo hello && sleep 100 && echo goodby'], stdout=stdout, iotimeout=1)
        self.assertNotEqual(p0.wait(), 0)
        self.assertIn('hello', stdout.getvalue())
        self.assertIn('Timeout: No output from program for 1 seconds', stdout.getvalue())
        self.assertFalse(p0._iostream.isAlive())


    @skipPosix
    def test_iotimeout_win(self):
        stdout = io.BytesIO()
        p0 = BufferedPopen(['cmd', '/c', 'echo hello && sleep 100 && echo goodby'], stdout=stdout, iotimeout=1)
        returncode = p0.wait()
        print("returncode", returncode)
        # self.assertNotEqual(returncode, 0)
        self.assertIn('hello', stdout.getvalue())
        self.assertIn('Timeout: No output from program for 1 seconds', stdout.getvalue())
        self.assertFalse(p0._io_thread.isAlive())


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.test_timeout']
    unittest.main()



from __future__ import (print_function, unicode_literals, division,
    absolute_import)

import atexit
from logging import getLogger
import shutil

from six.moves.urllib.parse import urlparse, urlunparse
from binstar_client import errors
import re


log = getLogger('binstar.git')

def is_url(path):
    url = urlparse(path)
    if url.scheme:
        return True

def get_urlpath(path):
    url = urlparse(path)
    return url.path[1:]  # path is /a/b (want to remove first slash)



def clone_repo(path):
    url = urlparse(path)
    fragment = url.fragment
    scheme = url.scheme[4:]

    git_url = urlunparse((scheme, url.netloc, url.path, '', '', ''))
    import tempfile
    from subprocess import check_call
    tmp_dir = tempfile.mkdtemp('.git', 'binstar-build')
    log.info(' '.join(['git', 'clone', git_url, tmp_dir]))
    check_call(['git', 'clone', git_url, tmp_dir])
    log.info(' '.join(['git', 'checkout', fragment]))
    check_call(['git', 'checkout', fragment], cwd=tmp_dir)

    def rmrepo():
        log.info("Removing git temp git repo")
        shutil.rmtree(tmp_dir, ignore_errors=True)

    atexit.register(rmrepo)

    return tmp_dir


def get_gitrepo(url):
    # split branch from repo

    if url.netloc != 'github.com':
        raise errors.UserError("Currently only github.com urls are supported (got %s)" % url.netloc)

    pat = re.compile('^/(?P<repo>[\w-]+/[\w-]+)(/tree/(?P<branch>[\w/]+))?(.git)?$')
    match = pat.match(url.path)
    if not match:
        raise errors.UserError("URL path '%s' is not a git repo" % url.path)

    groups = match.groupdict()
    repo = groups.get('repo')
    branch = groups.get('branch') or url.fragment or 'master'
    builds = {'repo': repo, 'branch':branch}
    return builds



'''
Trigger a build that has been saved

See also: 

  * [Save and Trigger Your Builds](http://docs.binstar.org/examples.html#SaveAndTriggerYourBuilds)
'''

from __future__ import (print_function, unicode_literals, division,
    absolute_import)

from argparse import RawDescriptionHelpFormatter
import logging

from binstar_build_client import BinstarBuildAPI
from binstar_client.utils import get_binstar
from binstar_client.utils import package_specs

log = logging.getLogger('binstar.build')

def main(args):

    binstar = get_binstar(args, cls=BinstarBuildAPI)

    build_no = binstar.trigger_build(args.package.user, args.package.name,
                               channels=args.channels, queue_name=args.queue,
                               branch=args.branch, test_only=args.test_only,
                               filter_platform=args.platform)

    log.info('')
    log.info('To view this build go to http://alpha.binstar.org/%s/%s/builds/matrix/%s' % (args.package.user, args.package.name, build_no))
    log.info('')
    log.info('You may also run\n\n    binstar-build tail -f %s/%s %s\n' % (args.package.user, args.package.name, build_no))
    log.info('')
    log.info('Build %s submitted' % build_no)


def add_parser(subparsers):
    description = 'Trigger a build that has been saved'
    parser = subparsers.add_parser('trigger',
                                      help=description,
                                      description=description,
                                      epilog=__doc__,
                                      formatter_class=RawDescriptionHelpFormatter,
                                      )

    parser.add_argument('package',
                       help="The binstar package to trigger a build on",
                       metavar='USER/PACKAGE',
                       type=package_specs)

    parser.add_argument('--channel', action='append', dest='channels',
                       help="Upload targets to this channel")

    parser.add_argument('--queue',
                       help="Build on this queue")

    parser.add_argument('--branch', default='master',
                        help="Branch to build"
                        )

    parser.add_argument('--platform',
                        help="Only run the build for this platform, (filters list from .binstar.yml)"
                        )

    parser.add_argument('--test-only', '--no-upload', action='store_true',
                        dest='test_only',
                        help="Don't upload the build targets to binstar, but run everything else")


    parser.set_defaults(main=main)




"""
Remove all conda build artifacts
"""

from __future__ import print_function

from argparse import ArgumentParser
import os

from binstar_build_client.utils import get_conda_root_prefix
from binstar_build_client.utils.rm import rm_rf


def main():
    parser = ArgumentParser(description=__doc__)
    parser.parse_args()

    root_env = get_conda_root_prefix()
    build_root = os.path.join(root_env, 'conda-bld')
    if os.path.isdir(build_root):
        print("Removing conda build root %s" % build_root)
        rm_rf(build_root)
    else:
        print("Conda build root %s does not exist" % build_root)


if __name__ == '__main__':
    main()



'''
Build command

Initialize the build directory:

    binstar build --init

This will create a default .binstar.yml file in the current directory

Submit a build:

    binstar build --submit

Tail the output of a build untill it is complete:

    binstar build --tail 1.0

'''

from __future__ import (print_function, unicode_literals, division,
    absolute_import)

import logging
import os
from os.path import abspath, join, basename
import sys

from binstar_build_client import BinstarBuildAPI
from binstar_build_client.utils.build_file import initial_build_config
from binstar_client import errors
from binstar_client.utils import get_binstar, bool_input
from six.moves import input


log = logging.getLogger('binstar.build')

def init_build(args):

    binstar = get_binstar(args, cls=BinstarBuildAPI)

    # Force user auth
    user = binstar.user()

    binstar_yml = join(args.path, '.binstar.yml')

    if os.path.exists(binstar_yml):
        result = bool_input("The file '%s' already exists. Would you like to overwrite it?" % binstar_yml,
                            default=False)
        if not result:
            log.error('goodby')
            sys.exit(1)

    name = basename(abspath(args.path))
    package_name = input('Please choose a name for this package: (default %s)\n> ' % name)
    package_name = package_name or name


    with open(binstar_yml, 'w') as fd:
        fd.write(initial_build_config % dict(PACKAGE_NAME=package_name))
    log.info("Wrote file '%s'" % binstar_yml)

    try:
        _ = binstar.package(user['login'], package_name)
    except errors.NotFound:
        log.warn('The package %(username)s/%(name)s does not exist\n'
                 'Please run:\n   binstar package %(username)s/%(name)s --create' % dict(username=user['login'], name=package_name))
    log.info("Run 'binstar-build submit' to submit your first build")
    return

def add_parser(subparsers):
    parser = subparsers.add_parser('init',
                                      help='Initialize Build file',
                                      description=__doc__,
                                      )
    parser.add_argument('path', default='.', nargs='?')
    parser.set_defaults(main=init_build)



'''
Binstar Build command

To get started with binstar build run:

    binstar-build init
    binstar-build submit

See also: 

  * [Binstar Build](http://docs.binstar.org/examples.html#BinstarBuild)
  
'''

from __future__ import (print_function, unicode_literals, division,
    absolute_import)

import logging

from binstar_build_client import __version__ as version
from binstar_build_client import commands
from binstar_client.scripts.cli import binstar_main


logger = logging.getLogger('binstar')

def main(args=None, exit=True):
    description = 'Binstar build client for continuous integration, testing and building packages'
    return binstar_main(commands, args, exit,
                        epilog=__doc__, description=description,
                        version=version)


if __name__ == '__main__':
    main()



'''
Build worker 
'''

from __future__ import (print_function, unicode_literals, division,
    absolute_import)

import logging

from binstar_build_client import BinstarBuildAPI
from binstar_build_client.worker.docker_worker import DockerWorker
from binstar_client.utils import get_binstar

from .worker import add_parser as add_worker_parser
from binstar_client import errors

try:
    import docker
except ImportError:
    docker = None

log = logging.getLogger('binstar.build')


def main(args):
    if docker is None:
        raise errors.UserError("binstar-build docker-worker requires docker and dockerpy to be installed\n"
                               "Run:\n\tpip install dockerpy")

    bs = get_binstar(args, cls=BinstarBuildAPI)

    if args.queue.count('/') == 1:
        username, queue = args.queue.split('/', 1)
        args.username = username
        args.queue = queue
    elif args.queue.count('-') == 2:
        _, username, queue = args.queue.split('-', 2)
        args.username = username
        args.queue = queue
    else:
        raise errors.UserError("Build queue must be of the form build-USERNAME-QUEUENAME or USERNAME/QUEUENAME")

    log.info('Starting worker:')
    log.info('User: %s' % args.username)
    log.info('Queue: %s' % args.queue)
    log.info('Platform: %s' % args.platform)
    woker = DockerWorker(bs, args)
    woker.work_forever()

def add_parser(subparsers):
    description = 'Run a build worker in a docker container to build jobs off of a binstar build queue'

    parser = add_worker_parser(subparsers, 'docker-worker',
                               description, __doc__)

    dgroup = parser.add_argument_group('docker arguments')
    dgroup.add_argument("-i", "--image", default="binstar/linux-64",
                        help="Docker image to use (default %(default)s)",
                        )
    dgroup.add_argument('--allow-user-images', action='store_true', default=False,
                        help="Allow user defined images")

    parser.set_defaults(main=main,
                        platform="linux-64",
                        conda_build_dir="/opt/miniconda/conda-bld/linux-64")





'''
Created on Feb 12, 2015

@author: sean
'''
import unittest
import mock
from binstar_build_client.utils import get_conda_root_prefix, CONDA_EXE
import os
class Test(unittest.TestCase):

    @mock.patch.dict(os.environ, {'PATH': '/does_not_exist!!'})
    def test_path_does_not_exist(self):
        prefix = get_conda_root_prefix()
        self.assertIsNone(prefix)

    @mock.patch.dict(os.environ, {'PATH': '/a/bin' + os.pathsep + '/b/bin'})
    @mock.patch('os.path.isdir')
    @mock.patch('os.listdir')
    def test_finds_conda(self, listdir, isdir):
        listdir.return_value = [CONDA_EXE, 'not_conda']
        prefix = get_conda_root_prefix()
        self.assertEqual(prefix, '/a')




if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
Save build info to be triggered later

See also: 

  * [Save and Trigger Your Builds](http://docs.binstar.org/examples.html#SaveAndTriggerYourBuilds)

'''

from __future__ import (print_function, unicode_literals, division,
    absolute_import)

from argparse import RawDescriptionHelpFormatter
import logging
import re

from binstar_build_client import BinstarBuildAPI
from binstar_client import errors
from binstar_client.utils import get_binstar, PackageSpec
from binstar_client.utils import package_specs
from six.moves.urllib.parse import urlparse
from binstar_build_client.utils.git_utils import get_gitrepo

log = logging.getLogger('binstar.build')

def main(args):

    binstar = get_binstar(args, cls=BinstarBuildAPI)

    # Force user auth

    package_name = None

    url = urlparse(args.url)
    builds = get_gitrepo(url)
    ghowner, ghrepo = builds['repo'].split('/', 1)

    if not args.package:
        package_name = ghrepo
        log.info("Using repo name '%s' as the pkg name." % package_name)
        user = binstar.user()
        user_name = user['login']
        args.package = PackageSpec(user_name, package_name)

    binstar = get_binstar(args, cls=BinstarBuildAPI)

    try:
        _ = binstar.package(args.package.user, args.package.name)
    except errors.NotFound:
        raise errors.UserError("Package %s does not exist" % (args.package,))

    log.info("Submitting the following repo for package creation: %s" % args.url)


    binstar.add_ci(args.package.user, args.package.name,
                   ghowner=ghowner, ghrepo=ghrepo,
                   channels=args.channels, queue=args.queue, sub_dir=args.sub_dir,
                   email=args.email)

    log.info("CI Added to package %s", args.package)

def add_parser(subparsers):

    description = 'Save build info to be triggered later'
    parser = subparsers.add_parser('save',
                                      help=description,
                                      description=description,
                                      epilog=__doc__,
                                      formatter_class=RawDescriptionHelpFormatter,
                                      )

    parser.add_argument('url',
                        help='The http github url to the repo')

    parser.add_argument('-p', '--package',
                       help="The binstar package namespace to upload the build to",
                       metavar='USER/PACKAGE',
                       type=package_specs)

    parser.add_argument('--sub-dir',
                       help="The sub directory within the git repository (github url submits only)")

    parser.add_argument('--channel', action='append', dest='channels',
                       help="Upload targets to this channel")

    parser.add_argument('--queue',
                       help="Build on this queue")

    parser.add_argument('--email', action='append',
                        help="Binstar usernames or email adresses to email when the build completes"
                        )

    parser.set_defaults(main=main)





from subprocess import Popen, PIPE, STDOUT
import unittest

from binstar_build_client.commands.worker import get_platform
from binstar_build_client.worker.utils.script_generator import gen_build_script
from os import path
import os

def default_build_data():
    return {
              'build_info':
                {'api_endpoint': 'api_endpoint',
                 'build_no': 1,
                 '_id':'build_id',
                 },
              'build_item_info':
                {'platform': get_platform(),
                 'engine': 'python',
                 'build_no': '1.0',
                 'sub_build_no': 0,
                 'instructions': {
                                  'install':'echo UNIQUE INSTALL MARKER',
                                  'test': 'echo UNIQUE TEST MARKER',
                                  'before_script': 'echo UNIQUE BEFORE SCRIPT MARKER',
                                  'script': 'echo UNIQUE SCRIPT MARKER',
                                  'after_failure': 'echo UNIQUE AFTER FAILURE MARKER',
                                  'after_error': 'echo UNIQUE AFTER ERROR MARKER',
                                  'after_success': 'echo UNIQUE AFTER SUCCESS MARKER',
                                  'after_script': 'echo UNIQUE AFTER SCRIPT MARKER',

                                  },
                 },
              'job':
                {'_id': 'test_gen_build_script'},
              'owner': {'login': 'me'},
              'package': {'name': 'the_package'},

              }

class Test(unittest.TestCase):

    def assertInOrdered(self, lst, container):
        container_orig = container
        while lst:
            item = lst.pop(0)
            index = container.find(item)
            if index < 0:
                if item in container_orig:
                    msg = "String %r is out of order in the given sequence" % (item)
                else:
                    msg = "String %r not found in output" % (item)
                assert False, msg
            container = container[index + len(item):]


    def test_bad_tarball(self):

        build_data = default_build_data()
        script_filename = gen_build_script(build_data, ignore_setup_build=True)
        self.addCleanup(os.unlink, script_filename)

        build_tarball = path.join(path.dirname(__file__), 'data', 'does_not_exist.tar.bz2')
        p0 = Popen([script_filename, '--build-tarball', build_tarball], stdout=PIPE, stderr=STDOUT)
        self.assertEqual(p0.wait(), 11)

    def test_instructions_success(self):

        build_data = default_build_data()
        script_filename = gen_build_script(build_data,
                                           ignore_setup_build=True,
                                           ignore_fetch_build_source=True)

        self.addCleanup(os.unlink, script_filename)
        p0 = Popen([script_filename], stdout=PIPE, stderr=STDOUT)
        return_code = p0.wait()
        self.assertEqual(return_code, 0,)
        output = p0.stdout.read()
        self.assertIn("Exit BINSTAR_BUILD_RESULT=success", output)

        self.assertInOrdered(['UNIQUE INSTALL MARKER',
                              'UNIQUE TEST MARKER',
                              'UNIQUE BEFORE SCRIPT MARKER',
                              'UNIQUE SCRIPT MARKER',
                              'UNIQUE AFTER SUCCESS MARKER',
                              'UNIQUE AFTER SCRIPT MARKER',
                              ], output)

    def test_instructions_error(self):

        build_data = default_build_data()
        build_data['build_item_info']['instructions']['install'] = 'invalid_command'
        script_filename = gen_build_script(build_data,
                                           ignore_setup_build=True,
                                           ignore_fetch_build_source=True)

        self.addCleanup(os.unlink, script_filename)
        p0 = Popen([script_filename], stdout=PIPE, stderr=STDOUT)
        return_code = p0.wait()
        output = p0.stdout.read()
        self.assertEqual(return_code, 11)

        self.assertIn("Exit BINSTAR_BUILD_RESULT=error", output)

        self.assertInOrdered([
                              'UNIQUE AFTER ERROR MARKER',
                              'UNIQUE AFTER SCRIPT MARKER',
                              ], output)

    def test_instructions_failure(self):

        build_data = default_build_data()
        build_data['build_item_info']['instructions']['test'] = 'invalid_command'
        script_filename = gen_build_script(build_data,
                                           ignore_setup_build=True,
                                           ignore_fetch_build_source=True)

        self.addCleanup(os.unlink, script_filename)
        p0 = Popen([script_filename], stdout=PIPE, stderr=STDOUT)
        return_code = p0.wait()
        output = p0.stdout.read()
        self.assertEqual(return_code, 12)

        self.assertIn("Exit BINSTAR_BUILD_RESULT=failure", output)

        self.assertInOrdered(['UNIQUE INSTALL MARKER',
                              'UNIQUE AFTER FAILURE MARKER',
                              'UNIQUE AFTER SCRIPT MARKER',
                              ], output)

    def test_instructions_failure2(self):

        build_data = default_build_data()
        build_data['build_item_info']['instructions']['script'] = 'invalid_command'
        script_filename = gen_build_script(build_data,
                                           ignore_setup_build=True,
                                           ignore_fetch_build_source=True)

        self.addCleanup(os.unlink, script_filename)
        p0 = Popen([script_filename], stdout=PIPE, stderr=STDOUT)
        return_code = p0.wait()
        output = p0.stdout.read()
        self.assertEqual(return_code, 12)

        self.assertIn("Exit BINSTAR_BUILD_RESULT=failure", output)

        self.assertInOrdered(['UNIQUE INSTALL MARKER',
                              'UNIQUE TEST MARKER',
                              'UNIQUE AFTER FAILURE MARKER',
                              'UNIQUE AFTER SCRIPT MARKER',
                              ], output)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.test_timeout']
    unittest.main()



"""
Threaded class to stream io from an iterator to a writeable file object
"""
import logging
from threading import Thread, Event
import time


log = logging.getLogger(__name__)


def stream_line_iterator(stream):
    """
    The builin file object does not work as an iterator in this case
    This is used as a wrapper 
    """
    data = stream.readline()
    while data:
        yield data
        data = stream.readline()


class IOStream(Thread):
    """
    Thread subclass 
    :param line_iterator: An iterable that yields output 
    :param outstream: The output to write to (must only have .write method)
    :param iotimeout: The duration in seconds to wait before calling `timeout_callback` if no output is yielded
    :param timeout: The duration in seconds to wait before calling `timeout_callback`
    :param timeout_callback: a function with the definition timeout_callback([True|False])
         called with true if `iotimeout` is reached otherwise `timeout` was reached
    """
    def __init__(self, line_iterator, outstream, iotimeout=None, timeout=None, timeout_callback=None):

        if isinstance(line_iterator, file):
            line_iterator = stream_line_iterator(line_iterator)

        self.line_iterator = line_iterator
        self.outstream = outstream
        self._start_time = self._last_io = time.time()
        self.iotimeout = iotimeout
        self.timeout = timeout
        self.timeout_callback = timeout_callback
        self.finished_event = Event()

        self._timeout_loop_thread = None
        Thread.__init__(self, None, name='iostream')

    def run(self):
        'Loop over lines in self.line_iterator and write them to outstream'
        try:

            for line in self.line_iterator:
                self._last_io = time.time()
                self.outstream.write(line)
                # self.outstream.flush()

                if getattr(self.outstream, 'terminate_build', False):
                    self.timeout_callback(reason='user_request')

        finally:
            self.finished_event.set()

    def start(self):

        if self.iotimeout or self.timeout:
            self._timeout_loop_thread = Thread(target=self.timeout_loop, name='timeout_loop')
            self._timeout_loop_thread.start()

        Thread.start(self)

    def timeout_loop(self):
        'Loop until finished_event is set'
        while not self.finished_event.wait(1):

            elapsed_io = time.time() - self._last_io

            if self.iotimeout and elapsed_io >= self.iotimeout:
                self.timeout_callback(reason='iotimeout')
                return

            elapsed_total = time.time() - self._start_time
            if self.timeout and elapsed_total >= self.timeout:
                self.timeout_callback(reason='timeout')
                return




from __future__ import (print_function, unicode_literals, division)

import imp
import os
MODULE_EXTENSIONS = ('.py', '.pyc', '.pyo')
import pkgutil
from os.path import dirname

def sub_command_names():
    return [name for _, name, _ in pkgutil.iter_modules([dirname(__file__)]) if not name.startswith('_')]

def sub_commands():
    names = sub_command_names()
    this_module = __import__(__package__, fromlist=names)
    for name in names:
        yield getattr(this_module, name)



from os.path import islink, isfile, isdir
import os
import sys
import shutil
import stat
import subprocess
import time

def rm_rf(path, max_retries=5):
    """
    Completely delete path
    max_retries is the number of times to retry on failure. The default is
    5. This only applies to deleting a directory.
    """
    if islink(path) or isfile(path):
        # Note that we have to check if the destination is a link because
        # exists('/path/to/dead-link') will return False, although
        # islink('/path/to/dead-link') is True.
        os.unlink(path)

    elif isdir(path):
        for i in range(max_retries):
            try:
                shutil.rmtree(path)
                return
            except OSError as e:
                msg = "Unable to delete %s\n%s\n" % (path, e)
                if sys.platform == 'win32':
                    try:
                        def remove_readonly(func, path, excinfo):
                            os.chmod(path, stat.S_IWRITE)
                            func(path)
                        shutil.rmtree(path, onerror=remove_readonly)
                        return
                    except OSError as e1:
                        msg += "Retry with onerror failed (%s)\n" % e1

                    try:
                        subprocess.check_call(['cmd', '/c', 'rd', '/s', '/q', path])
                        return
                    except subprocess.CalledProcessError as e2:
                        msg += '%s\n' % e2
                print(msg + "Retrying after %s seconds..." % i)
                time.sleep(i)

        # Final time. pass exceptions to caller.
        shutil.rmtree(path)



'''
Build command

Initialize the build directory:

    binstar build --init

This will create a default .binstar.yml file in the current directory

Submit a build:

    binstar build --submit

Tail the output of a build untill it is complete:

    binstar build --tail 1.0

'''


from __future__ import (print_function, unicode_literals, division,
    absolute_import)

from binstar_client.utils import get_binstar
from binstar_client.utils import package_specs
import argparse
import logging
from binstar_build_client import BinstarBuildAPI

log = logging.getLogger('binstar.build')

def halt_build(binstar, args):
    binstar.stop_build(args.package.user, args.package.name, args.halt)
    if args.halt == 'all':
        log.info('Stopping all builds')
    else:
        log.info('Stopping build %s' % args.halt)
    return

def resubmit_build(args):

    binstar = get_binstar(args, cls=BinstarBuildAPI)

    binstar.resubmit_build(args.package.user, args.package.name, args.build_no)

    log.info("Build %s resubmitted" % args.build_no)


def add_parser(subparsers):

    parser = subparsers.add_parser('resubmit',
                                      help='Resubmit build',
                                      description='Resubmit build',
                                      formatter_class=argparse.RawDescriptionHelpFormatter,
                                      )

    parser.add_argument('package', metavar='OWNER/PACKAGE',
                       help='build to the package OWNER/PACKAGE',
                       nargs='?',
                       type=package_specs)

    parser.add_argument('build_no',
                       help='The build to resubmit',
                       type=float)

    parser.set_defaults(main=resubmit_build)



'''
Attach a private key file
'''

from __future__ import (print_function, unicode_literals, division,
    absolute_import)

from binstar_client.utils import get_binstar
from binstar_client.utils import package_specs
import logging
from argparse import FileType
from binstar_build_client import BinstarBuildAPI

log = logging.getLogger('binstar.build')

def set_keyfile(args):
    binstar = get_binstar(args, cls=BinstarBuildAPI)
    if args.remove:
        binstar.remove_keyfile(args.package.user, args.package.name,
                               getattr(args, 'remote-filename'))
    else:
        content = args.upload.read()
        remote = getattr(args, 'remote-filename')
        binstar.set_keyfile(args.package.user, args.package.name,
                            remote, content)

def keyfiles(args):
    binstar = get_binstar(args, cls=BinstarBuildAPI)
    for key in binstar.keyfiles(args.package.user, args.package.name):
        print(key)


def add_parser(subparsers):
    description = '[Advanced] Not documented yet'
    parser = subparsers.add_parser('keyfiles',
                                      help=description, description=description,
                                      epilog=__doc__,
                                      )

    parser.add_argument('package', metavar='OWNER/PACKAGE',
                       help='build to the package OWNER/PACKAGE',
                       type=package_specs)

    parser.set_defaults(main=keyfiles)

    parser = subparsers.add_parser('keyfile',
                                      help=description, description=description,
                                      epilog=__doc__,
                                      )

    parser.add_argument('package', metavar='OWNER/PACKAGE',
                       help='build to the package OWNER/PACKAGE',
                       type=package_specs)

    parser.add_argument('remote-filename',
                       help='The filename on the build machine')

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-u', '--upload',
                       help='The file', type=FileType('r'))
    group.add_argument('-r', '--remove', action='store_true')

    parser.set_defaults(main=set_keyfile)



'''
Created on Feb 18, 2014

@author: sean
'''


initial_build_config = '''
## The package attribute specifies a binstar package namespace to build the package to. 
## This can be specified here or on the command line
package: %(PACKAGE_NAME)s

## You can also specify the account to upload to,
## you must be an admin of that account, this 
## defaults to your user account
# user: USERNAME 

#===============================================================================
# Build Matrix Options
# These options may be a single item, a list or empty
# The resulting number of builds is [platform * engine * env] 
#===============================================================================

## The platforms to build on.
## platform defaults to linux-64 
# platform: 
#  - linux-64
#  - linux-32
## The engine are the inital conda packages you want to run with 
# engine:
#  - python=2
#  - python=3
## The env param is an environment variable list 
# env:
#  - MY_ENV=A CC=gcc
#  - MY_ENV=B

#===============================================================================
# Script options
# These options may be broken out into the before_script, script and after_script
# or not, that is up to you 
#===============================================================================

## Run before the script
# before_script:
#   - echo "before_script!"
## Put your main computations here!  
script:
  - echo "This is my binstar build!"
## This will run after the script regardless of the result of script
## BINSTAR_BUILD_RESULT=[succcess|failure] 
# after_script:
#   - echo "The build was a $BINSTAR_BUILD_RESULT" | tee artifact1.txt
## This will be run only after a successful build
# after_success:
#   - echo "after_success!"
## This will be run only after a build failure
# after_failure:
#   - echo "after_failure!"

#===============================================================================
# Build Results
# Build results are split into two categories: artifacts and targets
# You may omit either key and stiff have a successful build
# They may be a string, list and contain any bash glob
#===============================================================================

## Build Targets: Upload these files to your binstar package
## build targets may be a list of files (globs allows) to upload
## The special build targets 'conda' and 'pypi' may be used to 
## upload conda builds  
## e.g. conda is an alias for /opt/anaconda/conda-bld/<os-arch>/*.tar.bz2
# build_targets:
#   - conda
'''



'''
Build Queue
'''
from __future__ import (print_function, unicode_literals, division,
    absolute_import)
from binstar_client.utils import get_binstar, bool_input
from binstar_build_client import BinstarBuildAPI
from argparse import RawDescriptionHelpFormatter
from dateutil.parser import parse as parse_date
from binstar_client.commands.authorizations import format_timedelta
from binstar_client import errors

def show_queue(queue):
    queue_owner = queue.get('owner')
    platforms = ', '.join(queue.get('platforms', []))

    queue_name = queue['_id']

    print('{queue_name:30} [{platforms}]'.format(**locals()))
    for worker in queue.get('workers', []):
        print(' + Worker hostname:{worker[hostname]:15} platform:{worker[platform]:15} dist:{worker[dist]:15}'.format(**locals()))

        try:
            last_seen = parse_date(worker['last_seen'])
            last_seen = '%s ago' % format_timedelta(last_seen, False)
        except TypeError:
            last_seen = worker['last_seen']

        print('   - Id {0[id]}'.format(worker))
        print('   - Last seen {0}'.format(last_seen))
        print('   - binstar-build v{0[binstar_build_version]} (binstar v{0[binstar_version]})'.format(worker))
    if not queue.get('workers', []):
        print(" + No build workers attached to this queue")
    print()

def show_queues(bs, username):
    print()
    for queue in bs.build_queues(username):
        show_queue(queue)

def main(args):
    bs = get_binstar(args, cls=BinstarBuildAPI)

    if args.queue is None:
        username = queue_name = None
    elif args.queue.count('/') == 1:
        username, queue_name = args.queue.split('/', 1)
    elif args.queue.count('/') == 2:
        _, username, queue_name = args.queue.split('/', 2)
    elif args.queue.count('-') == 2:
        _, username, queue_name = args.queue.split('-', 2)
    else:
        raise errors.UserError("Build queue must be of the form build-USERNAME-QUEUENAME or USERNAME/QUEUENAME")

    if args.create:

        if queue_name is None:
            raise errors.BinstarError("Must specify a queue name to create")
        bs.add_build_queue(username, queue_name)
        print("Created queue %s" % queue_name)
        return

    if queue_name:
        queue = bs.build_queue(username, queue_name)

    if args.remove:
        if queue.get('workers'):
            prompt = ('This build queue still has workers attached. '
                      'Are you sure you want to remove it')
            if not bool_input(prompt, False):
                print("Not removing queue")
                return
        bs.remove_build_queue(username, queue_name)
        print("Removed queue %s" % queue_name)
        return

    if args.remove_worker:
        bs.remove_worker(username, queue_name, args.remove_worker)
        print("Removed worker %s from queue %s" % (args.remove_worker, queue_name))
        return


    if queue_name:
        print()
        show_queue(queue)
    else:
        show_queues(bs, username)

def add_parser(subparsers):
    parser = subparsers.add_parser('queue',
                                      help='Inspect build queue',
                                      description=__doc__,
                                      formatter_class=RawDescriptionHelpFormatter,
                                      )

    parser.add_argument('queue', nargs='?', metavar='USERNAME/QUEUENAME',
                        help='Specify a queue to perform an operation on')
    parser.add_argument('-r', '--remove', action='store_true',
                        help='Remove the queue specified with the -q/--queue option')
    parser.add_argument('-c', '--create', action='store_true',
                        help='Create a new queue')
    parser.add_argument('--remove-worker', metavar='WORKER_ID',
                        help='Remove a worker from a queue')

    parser.set_defaults(main=main)



from __future__ import (print_function, unicode_literals, division,
    absolute_import)

import base64
import json
import os
import requests
import warnings

from binstar_build_client.mixins.build import BuildMixin
from binstar_build_client.mixins.build_queue import BuildQueueMixin
import logging
from binstar_client import Binstar

log = logging.getLogger('binstar.build')

try:
    from ._version import __version__
except ImportError:
    __version__ = '0.8'


class BinstarBuildAPI(BuildMixin, BuildQueueMixin, Binstar):
    '''
    '''
    pass



'''
Created on Feb 18, 2014

@author: sean
'''

from __future__ import (print_function, unicode_literals, division,
    absolute_import)

from mock import patch
import unittest

from binstar_build_client.scripts.build import main
from binstar_client.tests.fixture import CLITestCase
from binstar_client.tests.urlmock import urlpatch


class Test(CLITestCase):


    @urlpatch
    @patch('binstar_build_client.commands.worker.Worker')
    def test_worker_simple(self, urls, Worker):

        main(['--show-traceback', 'worker', 'username/queue-1'], False)
        self.assertEqual(Worker().work_forever.call_count, 1)

if __name__ == '__main__':
    unittest.main()



"""
Write IO back to build log on the binstar server
"""
import logging
import sys
from threading import Lock, Thread, Event
import traceback

from binstar_client import errors

from requests import ConnectionError

log = logging.getLogger(__name__)

class BuildLog(object):
    """
    This IO object writes data build log output to the binstar server and also to stdout
    
    This object first writes to a buffer that is sent to the server every BuildLog.INTERVAL 
    seconds
    """

    INTERVAL = 4

    def __init__(self, bs, username, queue, worker_id, job_id):
        self.bs = bs
        self.username = username
        self.queue = queue
        self.worker_id = worker_id
        self.job_id = job_id
        self._buffer = ''
        self._write_lock = Lock()
        self._running = False
        self.event = Event()


    def write(self, msg):
        """
        Write to server and also stdout
        
        The if the io thread is running, msg will be appended an internal message buffer
        """
        n = sys.stdout.write(msg)

        if self._running:
            with self._write_lock:
                self._buffer += msg
        else:
            terminate_build = self.bs.log_build_output(self.username, self.queue, self.worker_id, self.job_id, msg)
            self.terminate_build = terminate_build
        return n

    def __enter__(self):
        """
        Start a thread that will post to the server 
        """
        self._running = True
        self._io_thread = Thread(target=self._io_loop, name='io_loop')
        self._io_thread.start()
        self.terminate_build = False
        return self

    def __exit__(self, *args):

        if args[0] is not None:
            self.write("Build Error: An unhandled exception occurred in the build worker")
            self.write('---\n' + ''.join(traceback.format_exception(*args)) + '\n---')

        self._running = False
        self.event.set()
        self._io_thread.join()

    def flush(self):
        """
        Flush the current buffer to the server 
        """
        with self._write_lock:
            msg = self._buffer
            self._buffer = ''

        if not msg:
            return

        try:
            terminate_build = self.bs.log_build_output(self.username, self.queue, self.worker_id, self.job_id, msg)
            self.terminate_build = terminate_build
        except Exception as err:
            log.exception(err)
            # Insert data back to buffer for next write attempt
            with self._write_lock:
                self._buffer = msg + self._buffer


    def _io_loop(self):
        """
        Loop to write buffer to server 
        every self.INTERVAL seconds
        """
        while self._running:
            self.flush()
            self.event.wait(self.INTERVAL)

        self.flush()









import io
import os
import unittest

from binstar_build_client.commands.worker import get_platform
from binstar_build_client.worker.worker import Worker
from mock import Mock, patch


class MockWorker(Worker):
    def __init__(self):
        self.SLEEP_TIME = 0
        bs = Mock()
        args = Mock()
        args.hostname = 'test_hostname'
        args.platform = 'test_platform'
        args.timeout = 100
        Worker.__init__(self, bs, args)

def default_build_data():
    return {
              'build_info':
                {'api_endpoint': 'api_endpoint',
                 'build_no': 1,
                 '_id':'build_id',
                 },
              'build_item_info':
                {'platform': get_platform(),
                 'engine': 'python',
                 'build_no': '1.0',
                 'sub_build_no': 0,
                 'instructions': {
                                  'iotimeout': 61,
                                  'install':'echo UNIQUE INSTALL MARKER',
                                  'test': 'echo UNIQUE TEST MARKER',
                                  'before_script': 'echo UNIQUE BEFORE SCRIPT MARKER',
                                  'script': 'echo UNIQUE SCRIPT MARKER',
                                  'after_failure': 'echo UNIQUE AFTER FAILURE MARKER',
                                  'after_error': 'echo UNIQUE AFTER ERROR MARKER',
                                  'after_success': 'echo UNIQUE AFTER SUCCESS MARKER',
                                  'after_script': 'echo UNIQUE AFTER SCRIPT MARKER',

                                  },
                 },
              'job':
                {'_id': 'test_gen_build_script'},
              'owner': {'login': 'me'},
              'package': {'name': 'the_package'},
              'job_name':'job_name',
              'upload_token':'upload_token'
              }

class Test(unittest.TestCase):

    def test_worker_context(self):
        bs = Mock()
        bs.register_worker.return_value = 'test_worker_id'
        args = Mock()
        args.cwd = '.'

        worker = Worker(bs, args)
        worker.STATE_FILE = 'worker_test_state.yaml'

        with worker.worker_context() as worker_id:
            self.assertEqual(worker_id, 'test_worker_id')
            self.assertTrue(os.path.isfile(worker.STATE_FILE))

        self.assertEqual(bs.remove_worker.call_count, 1)
        self.assertFalse(os.path.isfile(worker.STATE_FILE))

    def test_handle_job(self):

        class MyWorker(MockWorker):
            build = Mock()
            build.return_value = False, 'success'
            download_build_source = Mock()

        worker = MyWorker()
        worker.args.push_back = False
        worker.worker_id = 'worker_id'
        worker._handle_job({'job':{'_id':'test_job_id'}})

        self.assertEqual(worker.build.call_count, 1)
        self.assertEqual(worker.bs.fininsh_build.call_count, 1)
        self.assertEqual(worker.bs.fininsh_build.call_args[1], {'status': 'success', 'failed': False})

    def test_failed_job(self):

        class MyWorker(MockWorker):
            build = Mock()
            build.side_effect = Exception

        worker = MyWorker()
        worker.args.push_back = False
        worker.worker_id = 'worker_id'

        worker._handle_job({'job':{'_id':'test_job_id'}})

        self.assertEqual(worker.build.call_count, 1)
        self.assertEqual(worker.bs.fininsh_build.call_count, 1)
        self.assertEqual(worker.bs.fininsh_build.call_args[1], {'status': 'error', 'failed': True})


    def test_download_build_source(self):

        worker = MockWorker()
        worker.worker_id = 'worker_id'
        expected = "build source"
        worker.bs.fetch_build_source.return_value = io.BytesIO(expected)

        filename = worker.download_build_source('job_id')
        self.assertTrue(os.path.isfile(filename))
        self.addCleanup(os.unlink, filename)

        with open(filename) as fd:
            data = fd.read()
        self.assertEqual(data, expected)




    @patch('binstar_build_client.worker.worker.BufferedPopen')
    @patch('binstar_build_client.worker.worker.gen_build_script')
    def test_build(self, gen_build_script, BufferedPopen):
        class MyWorker(MockWorker):
            download_build_source = Mock()
            download_build_source.return_value = 'build_source_filename'

        BufferedPopen.return_value.wait.return_value = 0
        gen_build_script.return_value = 'script_filename'

        worker = MyWorker()
        worker.worker_id = 'worker_id'
        job_data = default_build_data()

        failed, status = worker.build(job_data)
        self.assertFalse(failed)
        self.assertEqual(status, 'success')

        popen_args = BufferedPopen.call_args[0][0]
        expected_args = ['script_filename', '--api-token', 'upload_token', '--build-tarball', 'build_source_filename']
        self.assertTrue(popen_args[0].endswith('/script_filename'))
        self.assertEqual(popen_args[1:], expected_args[1:])
        popen_kwargs = BufferedPopen.call_args[1]
        self.assertEqual(popen_kwargs['iotimeout'], 61)

    def test_job_loop(self):
        worker = MockWorker()
        worker.worker_id = 'worker_id'
        worker.args.one = True
        worker.bs.pop_build_job.return_value = {'job':{'_id':'test_job_id'}, 'job_name':'job_name'}
        jobs = list(worker.job_loop())
        self.assertEqual(len(jobs), 1)

    def test_job_loop_error(self):

        worker = MockWorker()
        worker.worker_id = 'worker_id'
        worker.args.one = False
        worker.bs.pop_build_job.return_value = {'job':{'_id':'test_job_id'}, 'job_name':'job_name'}

        with self.assertRaises(TypeError):
            for job in worker.job_loop():
                raise TypeError()

    def test_job_context(self):

        worker = MockWorker()
        worker.worker_id = 'worker_id'
        worker.args.one = False
        job_data = {'job':{'_id':'test_job_id'}, 'job_name':'job_name'}
        journal = io.StringIO()

        with worker.job_context(journal, job_data):
            pass

        value = journal.getvalue()
        expected = 'starting build, test_job_id, job_name\nfinished build, test_job_id, job_name\n'
        self.assertEqual(value, expected)

    def test_job_context_error(self):

        worker = MockWorker()
        worker.worker_id = 'worker_id'
        worker.args.one = False
        job_data = {'job':{'_id':'test_job_id'}, 'job_name':'job_name'}
        journal = io.StringIO()

        with worker.job_context(journal, job_data):
            raise TypeError("hai")

        value = journal.getvalue()
        expected = 'starting build, test_job_id, job_name\nbuild errored, test_job_id, job_name\n'
        self.assertEqual(value, expected)

if __name__ == '__main__':
    unittest.main()



"""
The worker 
"""
from __future__ import print_function, absolute_import, unicode_literals

from contextlib import contextmanager
import logging
import os
import time

from binstar_build_client.utils.rm import rm_rf
from binstar_client import errors
import psutil
import requests
import yaml

from .utils.buffered_io import BufferedPopen
from .utils.build_log import BuildLog
from .utils.script_generator import gen_build_script, \
    EXIT_CODE_OK, EXIT_CODE_ERROR, EXIT_CODE_FAILED
import inspect


log = logging.getLogger('binstar.build')

def get_my_procs():

    this_proc = psutil.Process()

    if os.name == 'nt':
        myusername = this_proc.username()
        def ismyproc(proc):
            try:
                return proc.username() == myusername
            except psutil.AccessDenied:
                return False
    else:
        def ismyproc(proc):
            if inspect.isroutine(this_proc.uids):
                # psutil >= 2
                return proc.uids().real == this_proc.uids().real
            else:
                # psutil < 2
                return proc.uids.real == this_proc.uids.real


    return {proc.pid for proc in psutil.process_iter() if ismyproc(proc)}

@contextmanager
def remove_files_after(files):
    try:
        yield
    finally:
        for filename in files:
            if os.path.isfile(filename):
                os.unlink(filename)



class Worker(object):
    """
    
    """
    STATE_FILE = 'worker.yaml'
    JOURNAL_FILE = 'journal.csv'
    SLEEP_TIME = 10

    def __init__(self, bs, args):
        self.bs = bs
        self.args = args

    def work_forever(self):
        """
        Start a loop and continuously build forever
        """
        log.info('Working Forever')
        with self.worker_context() as worker_id:
            self.worker_id = worker_id
            self._build_loop()

    def job_loop(self):
        """
        An iterator that will yield job_data objects when
        one is available. 
        
        Also handles journaling of jobs
        
        """
        bs = self.bs
        args = self.args
        worker_idle = False
        while 1:
            try:
                job_data = bs.pop_build_job(args.username, args.queue, self.worker_id)
            except errors.NotFound:

                if args.show_traceback:
                    raise
                else:
                    msg = ("This worker can no longer pop items off the build queue. "
                           "Did someone remove it manually?")
                    raise errors.BinstarError(msg)

            except requests.ConnectionError as err:
                log.error("Trouble connecting to binstar at '%s' " % bs.domain)
                log.error("Could not retrieve work items")
                job_data = {}
            except errors.ServerError as err:
                log.exception(err)
                log.error("There server '%s' returned an error response " % bs.domain)
                log.error("Could not retrieve work items")
                job_data = {}

            if job_data.get('job') is None:
                if not worker_idle:
                    idle_msg = 'Worker is waiting for the next job'
                    log.info(idle_msg)
                worker_idle = True
                time.sleep(self.SLEEP_TIME)
                continue

            worker_idle = False

            yield job_data

            if args.one:
                break



    def _handle_job(self, job_data):
        """
        Handle a single build job
        only catches build script level errors
        """

        try:
            failed, status = self.build(job_data)
        except Exception as err:
            # Catch all exceptions here and submit a build error
            log.exception(err)
            failed = True
            status = 'error'
        except BaseException as err:
            # Catch all exceptions here and submit a build error
            log.exception(err)
            failed = True
            status = 'error'
            self._finish_job(job_data, failed, status)
            raise

        self._finish_job(job_data, failed, status)

    def _finish_job(self, job_data, failed, status):
        bs = self.bs
        args = self.args

        if args.push_back:
            bs.push_build_job(args.username, args.queue, self.worker_id, job_data['job']['_id'])
        else:
            job_data = bs.fininsh_build(args.username, args.queue, self.worker_id, job_data['job']['_id'],
                                        failed=failed, status=status)

    def _build_loop(self):
        """
        This is the main build loop this checks binstar.org for any jobs it can do and 
        """

        with open(self.JOURNAL_FILE, 'a') as journal:
            for job_data in self.job_loop():
                with self.job_context(journal, job_data):
                    self._handle_job(job_data)


    def build(self, job_data):
        """
        Run a single build 
        """
        job_id = job_data['job']['_id']
        with BuildLog(self.bs, self.args.username, self.args.queue, self.worker_id, job_id) as build_log:


            instructions = job_data['build_item_info'].get('instructions')
            build_log.write("Building on worker %s (platform %s)\n" % (self.args.hostname, self.args.platform))
            build_log.write("Starting build %s\n" % job_data['job_name'])

            if not os.path.exists('build_scripts'):
                os.mkdir('build_scripts')

            script_filename = gen_build_script(job_data,
                                               conda_build_dir=self.args.conda_build_dir)

            iotimeout = instructions.get('iotimeout', 60)
            timeout = self.args.timeout

            api_token = job_data['upload_token']

            files = [script_filename]

            git_oauth_token = job_data.get('git_oauth_token')
            if not job_data.get('build_info', {}).get('github_info'):
                build_filename = self.download_build_source(job_id)
                files.append(build_filename)
            else:
                build_filename = None

            with remove_files_after(files):
                exit_code = self.run(job_data, script_filename, build_log,
                                     timeout, iotimeout,
                                     api_token, git_oauth_token, build_filename,
                                     instructions=instructions)

            log.info("Build script exited with code %s" % exit_code)
            if exit_code == EXIT_CODE_OK:
                failed = False
                status = 'success'
                log.info('Build %s Succeeded' % (job_data['job_name']))
            elif exit_code == EXIT_CODE_ERROR:
                failed = True
                status = 'error'
                log.error("Build %s errored" % (job_data['job_name']))
            elif exit_code == EXIT_CODE_FAILED:
                failed = True
                status = 'failure'
                log.error("Build %s failed" % (job_data['job_name']))
            else:  # Unknown error
                failed = True
                status = 'error'
                log.error("Unknown build exit status %s for build %s" % (exit_code, job_data['job_name']))

            return failed, status

    def run(self, build_data, script_filename, build_log, timeout, iotimeout,
            api_token=None, git_oauth_token=None, build_filename=None, instructions=None):


        owner = build_data['owner']['login']
        package = build_data['package']['name']

        working_dir = os.path.abspath(os.path.join('builds', owner, package))
        rm_rf(working_dir)
        os.makedirs(working_dir)

        args = [os.path.abspath(script_filename), '--api-token', api_token]


        if git_oauth_token:
            args.extend(['--git-oauth-token', git_oauth_token])

        elif build_filename:
            args.extend(['--build-tarball', build_filename])

        log.info("Running command: (iotimeout=%s)" % iotimeout)
        log.info(" ".join(args))

        if self.args.show_new_procs:
            already_running_procs = get_my_procs()

        p0 = BufferedPopen(args, stdout=build_log, iotimeout=iotimeout, cwd=working_dir)

        try:
            exit_code = p0.wait()
        except BaseException:
            log.error("Binstar build process caught an exception while waiting for the build to finish")
            p0.kill_tree()
            p0.wait()
            raise
        finally:
            if self.args.show_new_procs:
                currently_running_procs = get_my_procs()
                new_procs = [psutil.Process(pid) for pid in currently_running_procs - already_running_procs]
                if new_procs:
                    build_log.write("WARNING: There are processes that were started during the build and are still running\n")
                    for proc in new_procs:
                        build_log.write(" - Process name:%s pid:%s\n" % (proc.name, proc.pid))
                        try:
                            cmdline = ' '.join(proc.cmdline)
                        except:
                            pass
                        else:
                            build_log.write("    + %s\n" % cmdline)


        return exit_code

    def download_build_source(self, job_id):
        """
        If the source files for this job were tarred and uploaded to bisntar.
        Download them. 
        """
        log.info("Fetching build data")
        if not os.path.exists('build_data'):
            os.mkdir('build_data')

        build_filename = os.path.join('build_data', '%s.tar.bz2' % job_id)
        fp = self.bs.fetch_build_source(self.args.username, self.args.queue, self.worker_id, job_id)

        with open(build_filename, 'wb') as bp:
            data = fp.read(2 ** 13)
            while data:
                bp.write(data)
                data = fp.read(2 ** 13)

        log.info("Wrote build data to %s" % build_filename)
        return os.path.abspath(build_filename)


    @contextmanager
    def worker_context(self):
        '''
        Register the worker with binstar and clean up on any excpetion or exit
        '''
        os.chdir(self.args.cwd)

        if os.path.isfile(self.STATE_FILE):
            with open(self.STATE_FILE, 'r') as fd:
                worker_data = yaml.load(fd)

            self.bs.remove_worker(self.args.username, self.args.queue, worker_data['worker_id'])
            log.info("Un-registered worker %s from binstar site" % worker_data['worker_id'])
            os.unlink(self.STATE_FILE)
            log.info("Removed worker.yaml")

        worker_id = self.bs.register_worker(self.args.username, self.args.queue, self.args.platform,
                                            self.args.hostname, self.args.dist)
        worker_data = {'worker_id': worker_id}

        with open(self.STATE_FILE, 'w') as fd:
            yaml.dump(worker_data, fd)
        try:
            yield worker_id
        finally:
            log.info("Removing worker %s" % worker_id)
            try:
                self.bs.remove_worker(self.args.username, self.args.queue, worker_id)
                os.unlink(self.STATE_FILE)
            except Exception as err:
                log.exception(err)
            log.debug("Removed %s" % self.STATE_FILE)

    @contextmanager
    def job_context(self, journal, job_data):
        """
        Yields a context where a job can execute safely
        
        If the context is not exited within 'args.timeout' seconds, an exception will be raised
        """
        ctx = (job_data['job']['_id'], job_data['job_name'])

        log.info('Starting build, %s, %s' % ctx)
        journal.write('starting build, %s, %s\n' % ctx)

        start_time = time.time()
        log.info('Setting alarm to terminate build after %i seconds' % self.args.timeout)

        try:
            yield
        except Exception as err:
            journal.write('build errored, %s, %s\n' % ctx)
            log.exception(err)
            time.sleep(self.SLEEP_TIME)
        else:
            journal.write('finished build, %s, %s\n' % ctx)
        finally:
            duration = time.time() - start_time
            log.info('Build Duration %i seconds' % duration)




'''
Created on May 8, 2014

@author: sean
'''


from __future__ import (print_function, unicode_literals, division,
    absolute_import)

from itertools import product

def expand_build_matrix(instruction_set):
    instruction_set = instruction_set.copy()

    platforms = instruction_set.pop('platform', ['linux-64']) or [None]
    if not isinstance(platforms, list): platforms = [platforms]
    envs = instruction_set.pop('env', [None]) or [None]
    if not isinstance(envs, list): envs = [envs]
    engines = instruction_set.pop('engine', ['python=2']) or [None]
    if not isinstance(engines, list): engines = [engines]

    for platform, env, engine in product(platforms, envs, engines):
        build = instruction_set.copy()
        build.update(platform=platform, env=env, engine=engine)
        yield build

def serialize_builds(instruction_sets):
    builds = {}
    for instruction_set in instruction_sets:
        for build in expand_build_matrix(instruction_set):
            k = '%s::%s::%s' % (build['platform'], build['engine'], build['env'])
            bld = builds.setdefault(k, build)
            bld.update(build)

    for k, value in sorted(builds.items()):
        if value.get('exclude'): continue
        yield value



'''
Build command

Submit a build from your local path or  via a git url:

See also: 

  * [Submit Your First Build](http://docs.binstar.org/examples.html#SubmitYourFirstBuild)
  * [Submit A Build From Github](http://docs.binstar.org/examples.html#SubmitABuildFromGithub)

'''

from __future__ import (print_function, unicode_literals, division,
    absolute_import)

from argparse import RawDescriptionHelpFormatter
from contextlib import contextmanager
import logging, yaml
import os
from os.path import abspath, join, isfile
import tarfile
import tempfile

from binstar_build_client import BinstarBuildAPI
from binstar_build_client.utils.filter import ExcludeGit
from binstar_build_client.utils.git_utils import is_url, get_urlpath, \
    get_gitrepo
from binstar_build_client.utils.matrix import serialize_builds
from binstar_client import errors
from binstar_client.errors import UserError
from binstar_client.utils import get_binstar, PackageSpec, upload_print_callback
from binstar_client.utils import package_specs
from six.moves.urllib.parse import urlparse

log = logging.getLogger('binstar.build')

@contextmanager
def mktemp(suffix=".tar.gz", prefix='binstar', dir=None):
    tmp = tempfile.mktemp(suffix, prefix, dir)
    log.debug('Creating temp file: %s' % tmp)
    try:
        yield tmp
    finally:
        log.debug('Removing temp file: %s' % tmp)
        os.unlink(tmp)

def submit_build(binstar, args):

    path = abspath(args.path)

    log.info('Getting build product: %s' % abspath(args.path))

    with open(join(path, '.binstar.yml')) as cfg:
        build_matrix = list(yaml.load_all(cfg))

    builds = list(serialize_builds(build_matrix))

    if args.platform:

        log.info("Only selecting builds on platform %s" % args.platform)
        builds = [b for b in builds if b['platform'] == args.platform]

    if not builds:
        msg = "No build instructions found"
        if args.platform:
            msg += " for platform %s" % args.platform
        raise errors.BinstarError(msg)

    log.info('Submitting %i sub builds' % len(builds))
    for i, build in enumerate(builds):
        log.info(' %i)' % i + ' %(platform)-10s  %(engine)-15s  %(env)-15s' % build)

    if not args.dry_run:
        if args.git_url:
            log.info("Submitting the following repo for package creation: %s" % args.git_url)

        else:
            with mktemp() as tmp:
                log.info("Archiving build directory for upload ...")
                with tarfile.open(tmp, mode='w|bz2') as tf:
                    exclude = ExcludeGit(path, use_git_ignore=not args.dont_git_ignore)
                    tf.add(path, '.', exclude=exclude)

                log.info("Created archive (%i files); Uploading to binstar" % exclude.num_included)
                queue_tags = []
                if args.buildhost:
                    queue_tags.append('hostname:%s' % args.buildhost)

                if args.dist:
                    queue_tags.append('dist:%s' % args.dist)

                with open(tmp, mode='rb') as fd:

                    build = binstar.submit_for_build(args.package.user, args.package.name, fd, builds,
                                                     channels=args.channels,
                                                     queue=args.queue, queue_tags=queue_tags,
                                                     test_only=args.test_only, callback=upload_print_callback(args))

                print_build_results(args, build)

    else:
        log.info('Build not submitted (dry-run)')

def print_build_results(args, build):

    log.info('')
    build_result_url = build.get('url')
    if not build_result_url:
        build_result_url = 'http://alpha.binstar.org/%s/%s/builds/matrix/%s' % (args.package.user, args.package.name, build['build_no'])
    log.info('To view this build go to %s' % build_result_url)
    log.info('')
    log.info('You may also run\n\n    binstar-build tail -f %s/%s %s\n' % (args.package.user, args.package.name, build['build_no']))
    log.info('')
    log.info('Build %s submitted' % build['build_no'])


def submit_git_build(binstar, args):

    log.info("Submitting the following repo for package creation: %s" % args.git_url)
    builds = get_gitrepo(urlparse(args.git_url))

    if not args.package:
        user = binstar.user()
        user_name = user['login']
        package_name = builds['repo'].split('/')[1]
        log.info("Using repo name '%s' as the pkg name." % package_name)
        args.package = PackageSpec(user_name, package_name)

    try:
        _ = binstar.package(args.package.user, args.package.name)
    except errors.NotFound:
        raise errors.UserError("Package %s does not exist" % (args.package,))


    if not args.dry_run:
        log.info("Submitting the following repo for package creation: %s" % args.git_url)
        builds = get_gitrepo(urlparse(args.path))
        build = binstar.submit_for_url_build(args.package.user, args.package.name, builds,
                                             channels=args.channels, queue=args.queue, sub_dir=args.sub_dir,
                                             test_only=args.test_only, callback=upload_print_callback(args),
                                             filter_platform=args.platform,
                                                )

        print_build_results(args, build)

    else:
        log.info('Build not submitted (dry-run)')


def main(args):

    binstar = get_binstar(args, cls=BinstarBuildAPI)

    package_name = None
    user_name = None

    if args.git_url:
        args.path = args.git_url

    if args.git_url or is_url(args.path):
        args.git_url = args.path
        args.git_url_path = get_urlpath(args.path)
        args.dont_git_ignore = True

        submit_git_build(binstar, args)


    # not a github repo (must check for valid .binstar.yml file
    else:
        binstar_yml = join(args.path, '.binstar.yml')

        if not isfile(binstar_yml):
            raise UserError("file %s does not exist\n perhaps you should run\n\n    binstar-build init\n" % binstar_yml)

        package_name = None
        user_name = None
        with open(binstar_yml) as cfg:
            for build in yaml.load_all(cfg):
                if build.get('package'):
                    package_name = build.get('package')
                if build.get('user'):
                    user_name = build.get('user')

        # Force package to exist
        if args.package:
            if user_name and not args.package.user == user_name:
                log.warn('User name does not match the user specified in the .binstar.yml file (%s != %s)', args.package.user, user_name)
            user_name = args.package.user
            if package_name and not args.package.name == package_name:
                log.warn('Package name does not match the user specified in the .binstar.yml file (%s != %s)', args.package.name, package_name)
            package_name = args.package.name
        else:
            if user_name is None:
                user = binstar.user()
                user_name = user['login']
            if not package_name:
                raise UserError("You must specify the package name in the .binstar.yml file or the command line")

        try:
            _ = binstar.package(user_name, package_name)
        except errors.NotFound:
            log.error("The package %s/%s does not exist." % (user_name, package_name))
            log.error("Run: \n\n    binstar package --create %s/%s\n\n to create this package" % (user_name, package_name))
            raise errors.NotFound('Package %s/%s' % (user_name, package_name))
        args.package = PackageSpec(user_name, package_name)

        submit_build(binstar, args)


def add_parser(subparsers):
    description = 'Submit a diectory or github repo for building'
    parser = subparsers.add_parser('submit',
                                   help=description, description=description,
                                   epilog=__doc__,
                                   formatter_class=RawDescriptionHelpFormatter,
                                   )

    parser.add_argument('path', default='.', nargs='?',
                        help='filepath or github url to submit')

    parser.add_argument('--git-url',
                       help="The github url with valid .binstar.yml file to clone")

    parser.add_argument('-n', '--dry-run',
                       help="Parse the build file but don't submit", action='store_true')

    parser.add_argument('--no-progress',
                       help="Don't show progress bar", action='store_true')

    parser.add_argument('--dont-git-ignore',
                       help="Don't ignore files from .gitignore", action='store_true')


    fgroup = parser.add_argument_group('filters')

    fgroup.add_argument('--buildhost',
                        help="The host name of the intended build worker")

    fgroup.add_argument('--dist',
                        help=("The os distribution of intended build worker (e.g centos, ubuntu) "
                              "Use 'binstar-build queue' to view the workers")
                        )

    fgroup.add_argument('--platform',
                        help=("The platform to run (e.g linux-64, win-64, osx-64, etc) "
                              "(default: all the platforms in the .binstar.yaml file)")
                        )

    cgroup = parser.add_argument_group('build control')
    parser.add_argument('--queue',
                       help="Build on this queue")

    cgroup.add_argument('--channel', action='append', dest='channels',
                       help="Upload targets to this channel")
    cgroup.add_argument('--test-only', '--no-upload', action='store_true',
                        dest='test_only',
                        help="Don't upload the build targets to binstar, but run everything else")
    cgroup.add_argument('-p', '--package',
                       help="The binstar package namespace to upload the build to",
                       metavar='USER/PACKAGE',
                       type=package_specs)

    cgroup.add_argument('--sub-dir',
                       help="The sub directory within the git repository (github url submits only)")

    parser.set_defaults(main=main)






import os

CONDA_EXE = 'conda.exe' if os.name == 'nt' else 'conda'

def get_conda_root_prefix():

    for entry in os.environ.get('PATH').split(os.pathsep):

        if os.path.isdir(entry) and CONDA_EXE in os.listdir(entry):
            conda_exe_path = os.path.realpath(os.path.join(entry, 'conda'))
            bin_dir = os.path.dirname(conda_exe_path)
            return os.path.dirname(bin_dir)




from binstar_client.utils import jencode
import requests
import binstar_client
import binstar_build_client

class BuildQueueMixin(object):

    def register_worker(self, username, queue_name, platform, hostname, dist):
        url = '%s/build-worker/%s/%s' % (self.domain, username, queue_name)
        data, headers = jencode(platform=platform, hostname=hostname, dist=dist,
                                binstar_version=binstar_client.__version__,
                                binstar_build_version=binstar_build_client.__version__)
        res = self.session.post(url, data=data, headers=headers)
        self._check_response(res, [200])
        return res.json()['worker_id']

    def remove_worker(self, username, queue_name, worker_id):
        '''Un-register a worker
        
        returns true if worker existed and was removed
        '''

        url = '%s/build-worker/%s/%s/%s' % (self.domain, username, queue_name, worker_id)
        res = self.session.delete(url)
        self._check_response(res, [200, 404])
        return res.status_code == 200

    def pop_build_job(self, username, queue_name, worker_id):
        '''Un-register a worker
        
        returns true if worker existed and was removed
        '''

        url = '%s/build-worker/%s/%s/%s/jobs' % (self.domain, username, queue_name, worker_id)
        res = self.session.post(url)
        self._check_response(res, [200])
        return res.json()

    def log_build_output(self, username, queue_name, worker_id, job_id, msg):
        url = '%s/build-worker/%s/%s/%s/jobs/%s/log' % (self.domain, username, queue_name, worker_id, job_id)
        res = self.session.post(url, data=msg)
        self._check_response(res, [201, 200])

        try:
            result = res.json().get('terminate_build', False)
        except ValueError:
            result = False

        return result

    def fininsh_build(self, username, queue_name, worker_id, job_id, status='success', failed=False):
        url = '%s/build-worker/%s/%s/%s/jobs/%s/finish' % (self.domain, username, queue_name, worker_id, job_id)
        data, headers = jencode(status=status, failed=failed)
        res = self.session.post(url, data=data, headers=headers)
        self._check_response(res, [200])
        return res.json()

    def push_build_job(self, username, queue_name, worker_id, job_id):
        url = '%s/build-worker/%s/%s/%s/jobs/%s/push' % (self.domain, username, queue_name, worker_id, job_id)
        res = self.session.post(url)
        self._check_response(res, [201])
        return

    def fetch_build_source(self, username, queue_name, worker_id, job_id):
        url = '%s/build-worker/%s/%s/%s/jobs/%s/build-source' % (self.domain, username, queue_name, worker_id, job_id)

        res = self.session.get(url, allow_redirects=False, stream=True)

        self._check_response(res, allowed=[302, 304, 200])

        if res.status_code == 304:
            return None
        elif res.status_code == 302:
            res = requests.get(res.headers['location'], stream=True, verify=True)

        return res.raw

    def build_queues(self, username=None):
        if username:
            url = '%s/build-queues/%s' % (self.domain, username)
        else:
            url = '%s/build-queues' % (self.domain)

        res = self.session.get(url)
        self._check_response(res)
        return res.json()

    def build_queue(self, username, queuename):
        url = '%s/build-queues/%s/%s' % (self.domain, username, queuename)

        res = self.session.get(url)
        self._check_response(res)
        return res.json()


    def remove_build_queue(self, username, queuename):

        url = '%s/build-queues/%s/%s' % (self.domain, username, queuename)
        res = self.session.delete(url)
        self._check_response(res, [201])
        return

    def add_build_queue(self, username, queuename):

        url = '%s/build-queues/%s/%s' % (self.domain, username, queuename)

        data, headers = jencode()
        res = self.session.post(url, data=data, headers=headers)

        self._check_response(res, [201])
        return




'''
Build command

Initialize the build directory:

    binstar build --init

This will create a default .binstar.yml file in the current directory

Submit a build:

    binstar build --submit

Tail the output of a build until it is complete:

    binstar build --tail 1.0

'''

from __future__ import (print_function, unicode_literals, division,
    absolute_import)

from binstar_client.utils import get_binstar, PackageSpec, bool_input
import logging, yaml
from os.path import abspath, join, isfile, basename
from binstar_client.errors import UserError
import tempfile
import tarfile
from contextlib import contextmanager
import os
from binstar_client.utils import package_specs
import time
from itertools import product
from binstar_client import errors
import argparse
from binstar_client.utils.build_file import initial_build_config
import sys

from six.moves import input

log = logging.getLogger('binstar.build')

@contextmanager
def mktemp(suffix=".tar.gz", prefix='binstar', dir=None):
    tmp = tempfile.mktemp(suffix, prefix, dir)
    log.debug('Creating temp file: %s' % tmp)
    try:
        yield tmp
    finally:
        log.debug('Removing temp file: %s' % tmp)
        os.unlink(tmp)


def tail(binstar, args):
    log_items = binstar.tail_build(args.package.user, args.package.name, args.tail, limit=args.n)
    for log_item in log_items['log']:
        log.info(log_item.get('msg'))

    last_entry = log_items['last_entry']

    while args.f and not log_items.get('finished'):
        time.sleep(4)
        log_items = binstar.tail_build(args.package.user, args.package.name, args.tail,
                                       after=last_entry)
        for log_item in log_items['log']:
            log.info(log_item.get('msg'))

        last_entry = log_items['last_entry'] or last_entry

    if log_items.get('finished'):
        if log_items['failed']:
            log.error('Build Failed')
        else:
            log.info('Build Succeeded')
    else:
        log.info('... Build still running ...')


def list_builds(binstar, args):

    log.info('Getting builds:')

    fmt = '%(build_no)15s | %(status)15s | %(platform)15s | %(engine)15s | %(env)15s'

    header = {'build_no':'Build #', 'status':'Status',
              'platform':'Platform',
              'engine':'Engine',
              'env':'Env',
              }

    log.info(fmt % header)

    log.info(fmt.replace('|', '+') % dict.fromkeys(header, '-' * 15))
    build_no = None if args.list is True else args.list
    for build in binstar.builds(args.package.user, args.package.name, build_no):
        for item in build['items']:
            item.setdefault('status', build.get('status', '?'))
            item['build_no'] = '%s.%s' % (build['build_no'], item['sub_build_no'])
            log.info(fmt % item)
    log.info('')
    return

def halt_build(binstar, args):
    binstar.stop_build(args.package.user, args.package.name, args.halt)
    if args.halt == 'all':
        log.info('Stopping all builds')
    else:
        log.info('Stopping build %s' % args.halt)
    return


def expand_build_matrix(instruction_set):
    instruction_set = instruction_set.copy()

    platforms = instruction_set.pop('platform', ['linux-64']) or [None]
    if not isinstance(platforms, list): platforms = [platforms]
    envs = instruction_set.pop('env', [None]) or [None]
    if not isinstance(envs, list): envs = [envs]
    engines = instruction_set.pop('engine', ['python=2']) or [None]
    if not isinstance(engines, list): engines = [engines]

    for platform, env, engine in product(platforms, envs, engines):
        build = instruction_set.copy()
        build.update(platform=platform, env=env, engine=engine)
        yield build

def serialize_builds(instruction_sets):
    builds = {}
    for instruction_set in instruction_sets:
        for build in expand_build_matrix(instruction_set):
            k = '%s::%s::%s' % (build['platform'], build['engine'], build['env'])
            bld = builds.setdefault(k, build)
            bld.update(build)

    for k, value in sorted(builds.items()):
        if value.get('exclude'): continue
        yield value


def submit_build(args):

    binstar = get_binstar(args)

    path = abspath(args.path)
    log.info('Getting build product: %s' % abspath(args.path))

    with open(join(path, '.binstar.yml')) as cfg:
        build_matrix = list(yaml.load_all(cfg))

    builds = list(serialize_builds(build_matrix))
    log.info('Submitting %i sub builds' % len(builds))
    for i, build in enumerate(builds):
        log.info(' %i)' % i + ' %(platform)-10s  %(engine)-15s  %(env)-15s' % build)

    if not args.dry_run:
        with mktemp() as tmp:
            with tarfile.open(tmp, mode='w|bz2') as tf:
                tf.add(path, '.')

            with open(tmp, mode='rb') as fd:

                build_no = binstar.submit_for_build(args.package.user, args.package.name, fd, builds)
        log.info('Build %s submitted' % build_no)
    else:
        log.info('Build not submitted (dry-run)')



def resubmit_build(binstar, args):
    binstar.resubmit_build(args.package.user, args.package.name, args.resubmit)




def init_build(args):

    binstar = get_binstar()

    # Force user auth
    user = binstar.user()

    binstar_yml = join(args.path, '.binstar.yml')

    if os.path.exists(binstar_yml):
        result = bool_input("The file '%s' already exists. Would you like to overwrite it?" % binstar_yml,
                            default=False)
        if not result:
            log.error('goodbye')
            sys.exit(1)

    name = basename(abspath(args.path))
    package_name = input('Please choose a name for this package: (default %s)\n> ' % name)
    package_name = package_name or name


    with open(binstar_yml, 'w') as fd:
        fd.write(initial_build_config % dict(PACKAGE_NAME=package_name))
    log.info("Wrote file '%s'" % binstar_yml)

    try:
        _ = binstar.package(user['login'], package_name)
    except errors.NotFound:
        log.warn('The package %(username)s/%(name)s does not exist\n'
                 'Please run:\n   binstar package %(username)s/%(name)s --create' % dict(username=user['login'], name=package_name))
    log.info("Run 'binstar build --submit' to submit your first build")
    return

def main(args):

    binstar = get_binstar()

    # Force user auth
    user = binstar.user()

    package_name = None
    user_name = None

    binstar_yml = join(args.path, '.binstar.yml')

    if not isfile(binstar_yml):
        raise UserError("file %s does not exist" % binstar_yml)

    with open(binstar_yml) as cfg:
        for build in yaml.load_all(cfg):
            package_name = build.get('package')
            user_name = build.get('user')

    # Force package to exist
    if args.package:
        if user_name and not args.package.user == user_name:
            log.warn('User name does not match the user specified in the .binstar.yml file (%s != %s)', args.package.user, user_name)
        user_name = args.package.user
        if package_name and not args.package.name == package_name:
            log.warn('Package name does not match the user specified in the .binstar.yml file (%s != %s)', args.package.name, package_name)
        package_name = args.package.name
    else:
        if user_name is None:
            user_name = user['login']
        if not package_name:
            raise UserError("You must specify the package name in the .binstar.yml file or the command line")

    try:
        _ = binstar.package(user_name, package_name)
    except errors.NotFound:
        log.error("The package %s/%s does not exist." % (user_name, package_name))
        log.error("Run: 'binstar package --create %s/%s' to create this package" % (user_name, package_name))
        raise errors.NotFound('Package %s/%s' % (user_name, package_name))
    args.package = PackageSpec(user_name, package_name)

    if args.resubmit:
        log.info("Re submit build %s", args.resubmit)
        return resubmit_build(binstar, args)

    if args.tail:
        return tail(binstar, args)

    if args.halt:
        return halt_build(binstar, args)

    if args.list:
        return list_builds(binstar, args)

def add_parser(subparsers):

    parser = subparsers.add_parser('build',
                                      help='Build command',
                                      description=__doc__,
                                      formatter_class=argparse.RawDescriptionHelpFormatter,
                                      )

    parser.add_argument('path', default='.', nargs='?')
    parser.add_argument('package', metavar='OWNER/PACKAGE',
                       help='build to the package OWNER/PACKAGE',
                       nargs='?',
                       type=package_specs)
    parser.add_argument('--test-only', '--no-upload', action='store_true',
                        help="Don't upload the build targets to binstar, but run everything else")

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-l', '--list', default=True, type=int,
                       help='List the sub builds for this package')
    group.add_argument('-a', '--list-all', action='store_true',
                       help='List all the builds and sub-builds for this package',
                       dest='list')
    group.add_argument('-t', '--tail', metavar='X.Y',
                       help='Tail the build output of build number X.Y')
    group.add_argument('-s', '--submit',
                       help='Submit the build', action='store_true')
    group.add_argument('--resubmit',
                       help='Re-submit an old sub build', type=float)
    group.add_argument('--dry-run',
                       help="Parse the build file but don't submit", action='store_true')


    group.add_argument('--halt', '--stop', metavar='build_id', dest='halt',
                       help='Stop a build for this package')
    group.add_argument('--halt-all', '--stop-all', action='store_const', const='all', dest='halt',
                       help='Stop all builds')


    group = parser.add_argument_group('Tail Options')
    group.add_argument('-n', metavar='#', type=int,
                       help='Number of lines for tail output')
    group.add_argument('-f', action='store_true',
                       help=('The -f option causes tail to not stop when end of current output is reached,'
                             ' but rather to wait for additional data to be appended to the input')
                       )

    parser.set_defaults(main=main)



"""

"""

import logging
import os
import pipes
import shlex
import jinja2

from binstar_build_client.utils import get_conda_root_prefix

try:
    unicode
except NameError:
    unicode = str

log = logging.getLogger(__name__)

#===============================================================================
# Script exit Codes
#===============================================================================
EXIT_CODE_OK = 0
EXIT_CODE_ERROR = 11
EXIT_CODE_FAILED = 12

#===============================================================================
# Helper functions
#===============================================================================

def get_channels(job_data):
    """
    Return channel string to pass to binstar upload
    """

    build_targets = job_data['build_item_info'].get('build_targets')

    # TODO use git branch
    branch = 'dev'.replace('/', ':')
    ctx = dict(branch=branch)

    if job_data['build_info'].get('channels'):
        channels = job_data['build_info'].get('channels')
    elif isinstance(build_targets, dict):
        channels = build_targets.get('channels', [branch])
    else:
        channels = [branch]

    if not isinstance(channels, list): channels = [channels]
    _channels = []

    for ch in channels:
        try: _channels.append(ch % ctx)
        except (KeyError, ValueError):
            log.info('Bad channel value %r' % ch)

    channels = ' --channel ' + ' --channel '.join(_channels) if _channels else ''
    return channels


def get_files(context, job_data):
    """
    Return a list of files to run binstar upload on
    """
    build_targets = job_data['build_item_info'].get('instructions', {}).get('build_targets')
    if not build_targets:
        return []

    if isinstance(build_targets, basestring):
        build_targets = [build_targets]
    elif isinstance(build_targets, dict):
        build_targets = get_list(build_targets, 'files', default=[])

    if 'conda' in build_targets:
        idx = build_targets.index('conda')
        conda_build_dir = context['conda_build_dir']
        build_targets[idx] = os.path.join(conda_build_dir, '*.tar.bz2')

    if 'pypi' in build_targets:
        idx = build_targets.index('pypi')
        build_targets[idx] = 'dist/*'

    return build_targets

def get_list(dct, item, default=()):
    """
    Get an item from a dictionary, like `dict.get`. 
    
    This method will transform all scalar values into lists of lenght 1
    """
    value = dct.get(item, default)
    if not isinstance(value, (list, tuple)): value = [value]
    return list(value)

def create_git_context(build):
    """
    Create the git_info object for git source builds
    """
    git_info = {}
    github_info = build.get('github_info', {})
    if github_info:
        ghrepo = github_info['repository']
        ghowner = ghrepo['owner'].get('name', ghrepo['owner'].get('login'))
        git_info['full_name'] = '%s/%s' % (ghowner, ghrepo['name'])
        git_info['branch'] = github_info['ref'].split('/', 2)[-1]
        git_info['commit'] = github_info['after']
    return git_info

def create_exports(build_data):
    """
    Create a dict of environment variables for the build script
    """
    conda_root_prefix = get_conda_root_prefix()
    build_item = build_data['build_item_info']
    build = build_data['build_info']

    api_site = build['api_endpoint']

    exports = {
            # The build number as MAJOR.MINOR
            'BINSTAR_BUILD': build_item['build_no'],
            'BINSTAR_BUILD_MAJOR': build['build_no'],
            'BINSTAR_BUILD_MINOR': build_item['sub_build_no'],
            # the engine from the engine tag
            'BINSTAR_ENGINE': build_item.get('engine'),
            # the platform from the platform tag
            'BINSTAR_PLATFORM': build_item.get('platform', 'linux-64'),
            'BINSTAR_API_SITE': api_site,
            'BINSTAR_OWNER': build_data['owner']['login'],
            'BINSTAR_PACKAGE': build_data['package']['name'],
            'BINSTAR_BUILD_ID': build['_id'],
            'CONDA_BUILD_DIR': os.path.join(conda_root_prefix, 'conda-bld', build_item.get('platform', 'linux-64')),
            'BUILD_BASE': 'builds',
            'BUILD_ENV_DIR': 'build_envs',
           }


    build_env = build_item.get('env')
    if isinstance(build_env, (str, unicode)):
        _build_env = {}
        for item in shlex.split(build_env):
            if '=' in item:
                key, value = item.split('=', 1)
                _build_env[key] = value

        build_env = _build_env

    if isinstance(build_env, dict):
        exports.update(build_env)

    return exports

#===============================================================================
# Generate
#===============================================================================

def gen_build_script(build_data, **context):
    """
    Generate a build script from a submitted build
    
    :return: the filename of the build script to execute
    """

    platform = build_data['build_item_info']['platform']
    job_id = build_data['job']['_id']

    env = jinja2.Environment(loader=jinja2.PackageLoader(__name__, 'data'))
    env.globals.update(get_list=get_list, quote=lambda item: pipes.quote(str(item)))

    if platform in ['win-32', 'win-64']:
        build_script_template = env.get_or_select_template('build_script.bat')
        script_filename = os.path.join('build_scripts', '%s.bat' % job_id)
    else:
        build_script_template = env.get_or_select_template('build_script.sh')
        script_filename = os.path.join('build_scripts', '%s.sh' % job_id)


    exports = create_exports(build_data)

    context.update({'exports': sorted(exports.items()),
                    'instructions': build_data['build_item_info'].get('instructions', {}),
                    'git_info': create_git_context(build_data['build_info']),
                    'test_only': build_data['build_info'].get('test_only', False),
                    'sub_dir': build_data['build_info'].get('sub_dir'),
                    'channels': get_channels(build_data),
                    'files': get_files(context, build_data),
                    'EXIT_CODE_OK': 0,
                    'EXIT_CODE_ERROR': 11,
                    'EXIT_CODE_FAILED': 12,
               })



    build_script = build_script_template.render(**context)

    if not os.path.isdir('build_scripts'):
        os.mkdir('build_scripts')

    with open(script_filename, 'w') as fd:
        fd.write(build_script)

    if os.name != 'nt':
        os.chmod(script_filename, 0o777)

    return script_filename




'''
Build worker 
'''

from __future__ import (print_function, unicode_literals, division,
    absolute_import)

import logging
import platform

from binstar_build_client import BinstarBuildAPI
from binstar_build_client.worker.worker import Worker
from binstar_client.utils import get_binstar
import os
from binstar_build_client.utils import get_conda_root_prefix
from binstar_client import errors


log = logging.getLogger('binstar.build')


def main(args):

    args.conda_build_dir = args.conda_build_dir.format(args=args)
    bs = get_binstar(args, cls=BinstarBuildAPI)

    if args.queue.count('/') == 1:
        username, queue = args.queue.split('/', 1)
        args.username = username
        args.queue = queue
    elif args.queue.count('-') == 2:
        _, username, queue = args.queue.split('-', 2)
        args.username = username
        args.queue = queue
    else:
        raise errors.UserError("Build queue must be of the form build-USERNAME-QUEUENAME or USERNAME/QUEUENAME")

    log.info('Starting worker:')
    log.info('User: %s' % args.username)
    log.info('Queue: %s' % args.queue)
    log.info('Platform: %s' % args.platform)
    woker = Worker(bs, args)
    woker.work_forever()


OS_MAP = {'darwin': 'osx', 'windows':'win'}
ARCH_MAP = {'x86': '32',
            'i686': '32',
            'x86_64': '64',
			'amd64' : '64',
            }

def get_platform():
    operating_system = platform.system().lower()
    arch = platform.machine().lower()
    return '%s-%s' % (OS_MAP.get(operating_system, operating_system),
                      ARCH_MAP.get(arch, arch))

def get_dist():
    if platform.dist()[0]:
        return platform.dist()[0].lower()
    elif platform.mac_ver()[0]:
        darwin_version = platform.mac_ver()[0].rsplit('.', 1)[0]
        return 'darwin%s' % darwin_version
    elif platform.win32_ver()[0]:
        return platform.win32_ver()[0].lower()
    return 'unknown'


def add_parser(subparsers, name='worker',
               description='Run a build worker to build jobs off of a binstar build queue',
               epilog=__doc__):

    parser = subparsers.add_parser(name,
                                   help=description, description=description,
                                   epilog=epilog
                                   )

    conda_platform = get_platform()
    parser.add_argument('queue', metavar='OWNER/QUEUE',
                        help='The queue to pull builds from')
    parser.add_argument('-p', '--platform',
                        default=conda_platform,
                        help='The platform this worker is running on (default: %(default)s)')

    parser.add_argument('--hostname', default=platform.node(),
                        help='The host name the worker should use (default: %(default)s)')

    parser.add_argument('--dist', default=get_dist(),
                        help='The operating system distribution the worker should use (default: %(default)s)')

    parser.add_argument('--cwd', default='.',
                        help='The root directory this build should use (default: "%(default)s")')
    parser.add_argument('-t', '--max-job-duration', type=int, metavar='SECONDS',
                        dest='timeout',
                        help='Force jobs to stop after they exceed duration (default: %(default)s)', default=60 * 60 * 60)

    dgroup = parser.add_argument_group('development options')

    dgroup.add_argument("--conda-build-dir",
                        default=os.path.join(get_conda_root_prefix(), 'conda-bld', '{args.platform}'),
                        help="[Advanced] The conda build directory (default: %(default)s)",
                        )
    dgroup.add_argument('--show-new-procs', action='store_true', dest='show_new_procs',
                        help='Print any process that started during the build '
                             'and is still running after the build finished')

    dgroup.add_argument('-c', '--clean', action='store_true',
                        help='Clean up an existing workers session')
    dgroup.add_argument('-f', '--fail', action='store_true',
                        help='Exit main loop on any un-handled exception')
    dgroup.add_argument('-1', '--one', action='store_true',
                        help='Exit main loop after only one build')
    dgroup.add_argument('--push-back', action='store_true',
                        help='Developers only, always push the build *back* onto the build queue')

    parser.set_defaults(main=main)

    return parser




from setuptools import setup, find_packages

ctx = {}
try:
    with open('binstar_build_client/_version.py') as fd:
        exec(open('binstar_build_client/_version.py').read(), ctx)
    version = ctx.get('__version__', 'dev')
except IOError:
    version = '0.8'

setup(
    name='binstar-build',
    version=version,
    author='Sean Ross-Ross',
    author_email='srossross@gmail.com',
    url='http://github.com/Binstar/binstar_client',
    packages=find_packages(),
    install_requires=['binstar',
                      'jinja2', 'psutil'],

    include_package_data=True,
    zip_safe=False,

    entry_points={
          'console_scripts': [
              'binstar-build = binstar_build_client.scripts.build:main',
              'conda-clean-build-dir = binstar_build_client.scripts.conda_clean_build_dir:main',
              ]
                 },

)



'''
Build command

Initialize the build directory:

    binstar build --init

This will create a default .binstar.yml file in the current directory

Submit a build:

    binstar build --submit

Tail the output of a build until it is complete:

    binstar build --tail 1.0

'''

from __future__ import (print_function, unicode_literals, division,
    absolute_import)

from binstar_client.utils import get_binstar
from binstar_client.utils import package_specs
import time
import logging
from binstar_build_client import BinstarBuildAPI

log = logging.getLogger('binstar.build')

def tail(args):

    binstar = get_binstar(args, cls=BinstarBuildAPI)

    log_items = binstar.tail_build(args.package.user, args.package.name, args.build_no, limit=args.n)
    for log_item in log_items['log']:
        log.info(log_item.get('msg'))

    last_entry = log_items['last_entry']

    while args.f and not log_items.get('finished'):
        time.sleep(4)
        log_items = binstar.tail_build(args.package.user, args.package.name, args.build_no,
                                       after=last_entry)
        for log_item in log_items['log']:
            log.info(log_item.get('msg'))

        last_entry = log_items['last_entry'] or last_entry

    if log_items.get('finished'):
        if log_items['failed']:
            log.error('Build Failed')
        else:
            log.info('Build Succeeded')
    else:
        log.info('... Build still running ...')


def list_builds(args):

    binstar = get_binstar(args, cls=BinstarBuildAPI)

    log.info('Getting builds:')

    fmt = '%(build_no)15s | %(status)15s | %(platform)15s | %(engine)15s | %(env)15s'

    header = {'build_no':'Build #', 'status':'Status',
              'platform':'Platform',
              'engine':'Engine',
              'env':'Env',
              }

    log.info(fmt % header)

    log.info(fmt.replace('|', '+') % dict.fromkeys(header, '-' * 15))
    for build in binstar.builds(args.package.user, args.package.name, args.build_no):
        for item in build['items']:
            item.setdefault('status', build.get('status', '?'))
            item['build_no'] = '%s.%s' % (build['build_no'], item['sub_build_no'])
            log.info(fmt % item)
    log.info('')
    return


def add_parser(subparsers):
    parser = subparsers.add_parser('tail',
                                      help='Tail the build output of build number X.Y',
                                      description=__doc__,
                                      )

    parser.add_argument('package', metavar='OWNER/PACKAGE',
                       help='build to the package OWNER/PACKAGE',
                       type=package_specs)

    parser.add_argument('build_no',
                       help='Tail the build output of build number X.Y',
                       type=float)

    group = parser.add_argument_group('Tail Options')
    group.add_argument('-n', metavar='#', type=int,
                       help='Number of lines for tail output')

    group.add_argument('-f', action='store_true',
                       help=('The -f option causes tail to not stop when end of current output is reached,'
                             ' but rather to wait for additional data to be appended to the input')
                       )

    parser.set_defaults(main=tail)
    #===========================================================================
    #
    #===========================================================================
    parser = subparsers.add_parser('list-all',
                                      help='list the builds for package',
                                      description=__doc__,
                                      )
    parser.add_argument('package', metavar='OWNER/PACKAGE',
                       help='build to the package OWNER/PACKAGE',
                       nargs='?',
                       type=package_specs)
    parser.set_defaults(main=list_builds, build_no=None)
    #===========================================================================
    #
    #===========================================================================
    parser = subparsers.add_parser('list',
                                      help='list the builds for package',
                                      description=__doc__,
                                      )
    parser.add_argument('package', metavar='OWNER/PACKAGE',
                       help='build to the package OWNER/PACKAGE',
                       type=package_specs)

    parser.add_argument('build_no',
                       help='Tail the build output of build number X.Y',
                       type=int)
    parser.set_defaults(main=list_builds)



from __future__ import print_function
import logging
import os
from os.path import basename, abspath

from binstar_build_client.worker.utils.streamio import IOStream
from binstar_build_client.worker.worker import Worker
from requests import ConnectionError
from binstar_client import errors
import json

log = logging.getLogger("binstar.build")

try:
    import docker
except ImportError:
    docker = None


class DockerWorker(Worker):
    """
    """
    def __init__(self, bs, args):
        Worker.__init__(self, bs, args)

        self.client = docker.Client(base_url=os.environ.get('DOCKER_HOST'))
        log.info('Connecting to docker daemon ...')
        try:
            images = self.client.images(args.image)
        except ConnectionError as err:
            raise errors.BinstarError("Docker client could not connect to daemon (is docker installed?)\n"
                                      "You may need to set your DOCKER_HOST environment variable")
        if not images:
            raise errors.BinstarError("You do not have the docker image '%(image)s'\n"
                                      "You may need to run:\n\n\tdocker pull %(image)s\n" % dict(image=args.image))

        if self.args.allow_user_images:
            log.warn("Allowing users to specify docker images")


    def run(self, build_data, script_filename, build_log, timeout, iotimeout,
            api_token=None, git_oauth_token=None, build_filename=None, instructions=None):
        """
        """
        cli = self.client
        image = self.args.image
        container_script_filename = '/%s' % basename(script_filename)

        volumes = [container_script_filename,
                   ]
        binds = {abspath(script_filename): {'bind': container_script_filename, 'ro': False}}

        args = ["bash", container_script_filename, '--api-token', api_token]

        if git_oauth_token:
            args.extend(['--git-oauth-token', git_oauth_token])

        elif build_filename:
            container_build_filename = '/%s' % basename(build_filename)
            volumes.append(container_build_filename)
            binds[build_filename] = {'bind': container_build_filename, 'ro': False}
            args.extend(['--build-tarball', container_build_filename])

        log.info("Running command: (iotimeout=%s)" % iotimeout)
        if self.args.allow_user_images:
            if instructions and instructions.get('docker_image'):
                image = instructions['docker_image']
                if ':' in image:
                    repository, tag = image.rsplit(':', 1)
                else:
                    repository, tag = image, None

                build_log.write('Docker: Pull %s\n' % image)
                for line in cli.pull(repository, tag=tag, stream=True):
                    msg = json.loads(line)
                    if msg.get('status') == 'Downloading':
                        build_log.write('.')
                    elif msg.get('status'):
                        build_log.write(msg.get('status', '') + '\n')
                    else:
                        build_log.write(line + '\n')

        else:
            if instructions and instructions.get('docker_image'):
                build_log.write("WARNING: User specified images are not allowed on this build worker\n")
                build_log.write("Using default docker image\n")

        command = " ".join(args)
        log.info(command)
        build_log.write("Docker Image: %s\n" % image)
        log.info("Volumes: %r" % volumes)

        build_log.write("Docker: Create container\n")
        cont = cli.create_container(image, command=command, volumes=volumes)

        build_log.write("Docker: Attach output\n")
        stream = cli.attach(cont, stream=True, stdout=True, stderr=True)

        def timeout_callback(reason='iotimeout'):

            cli.kill(cont)

            if reason == 'iotimeout':
                build_log.write("\nTimeout: No output from program for %s seconds\n" % iotimeout)
                build_log.write("\nTimeout: If you require a longer timeout you "
                          "may set the 'iotimeout' variable in your .binstar.yml file\n")
                self._output.write("[Terminating]\n")
            elif reason == 'timeout':
                build_log.write("\nTimeout: build exceeded maximum build time of %s seconds\n" % timeout)
                build_log.write("[Terminating]\n")
            else:
                build_log.write("\nTerminate: User requested build to be terminated\n")
                build_log.write("[Terminating]\n")


        ios = IOStream(stream, build_log, iotimeout, timeout, timeout_callback)

        build_log.write("Docker: Start\n")

        ios.start()

        log.info("Binds: %r" % binds)

        cli.start(cont, binds=binds)

        exit_code = cli.wait(cont)

        ios.join()

        log.info("Remove Container: %r" % cont)
        cli.remove_container(cont, v=True)

        return exit_code







from __future__ import (print_function, unicode_literals, division,
    absolute_import)

import os
from subprocess import check_output, CalledProcessError

class ExcludeGit(object):
    def __init__(self, path, use_git_ignore=True):
        self.path = os.path.abspath(path)
        try:
            filelist = check_output(['git', 'ls-files'], cwd=self.path).decode().split()
            self.to_include = [os.path.join(self.path, fn) for fn in filelist]
        except CalledProcessError as err:
            self.to_include = None

        self.num_included = 0

    def __call__(self, filename):
        if self.to_include is None:
            return False

        if os.path.isdir(filename):
            return False

        if filename in self.to_include:
            self.num_included += 1
            return False

        return True



"""
Popen module with buffered IO 

i.e. stdout, stderr can be any file 'like' object. Like an io.BytesIO() object

Also this adds a new keyword argument iotimeout which will terminate the process if no output is recieved for 
iotimeout seconds  
"""

from __future__ import print_function

import logging
from subprocess import Popen, STDOUT, PIPE
import psutil
from binstar_build_client.worker.utils.streamio import IOStream



log = logging.getLogger(__name__)

def is_special(fd):
    """
    Test if a stream argument to POPEN is a subprocess.STDOUT or  subprocess.PIPE etc.
    """

    if fd is None:
        return True
    elif isinstance(fd, int) and fd <= 0:
        return True
    return False

class BufferedPopen(Popen):
    """
    Open a process and Buffer the output to *any* IO object (like `io.BytesIO`)
    """
    def __init__(self, args, stdout=None, iotimeout=None, timeout=None, **kwargs):
        self._output = stdout

        Popen.__init__(self, args, stdout=PIPE, stderr=STDOUT,
                       **kwargs)

        self._iostream = IOStream(self.stdout, self._output, iotimeout, timeout, self.timeout_callback)
        self._iostream.start()

    def wait(self):
        """Wait for child process to terminate.  Returns returncode
        attribute.
        
        If timeout is given, the process will be killed after timeout seconds if it is not finished 
        """
        returncode = Popen.wait(self)

#         self._finished_event.set()
        log.debug("returncode", returncode)

        if self._iostream.is_alive():
            log.debug("self._io_thread.join()")
            self._iostream.join()

        return returncode

    def timeout_callback(self, reason='iotimeout'):

        log.debug("timeout_callback")

        if reason == 'iotimeout':
            self._output.write("\nTimeout: No output from program for %s seconds\n" % self._iostream.iotimeout)
            self._output.write("\nTimeout: If you require a longer timeout you "
                      "may set the 'iotimeout' variable in your .binstar.yml file\n")
            self._output.write("[Terminating]\n")
        elif reason == 'timeout':
            self._output.write("\nTimeout: build exceeded maximum build time of %s seconds\n" % self._iostream.timeout)
            self._output.write("[Terminating]\n")
        else:
            self._output.write("\nTerminate: User requested build to be terminated\n")
            self._output.write("[Terminating]\n")

        self.kill_tree()

    def kill_tree(self):
        'Kill all processes and child processes'
        try:
            log.warn("Kill Tree parent pid:%s" % self.pid)
            parent = psutil.Process(self.pid)
        except psutil.NoSuchProcess:
            log.warn("Parent pid %s is already dead" % self.pid)
            # Already dead
            return

        children = parent.get_children(recursive=True)

        self.kill()
        for child in children:
            if child.is_running():
                log.warn(" - Kill child pid %s" % child.pid)
                child.kill()

